<?php
/**
 * @package   DPCalendar
 * @copyright Copyright (C) 2014 Digital Peak GmbH. <https://www.digital-peak.com>
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 */

defined('_JEXEC') or die();

if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
	return;
}

use DPCalendar\Helper\ExportTrait;
use DPCalendar\Plugin\SyncPlugin;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Date\Date;
use Joomla\CMS\Toolbar\Toolbar;
use Joomla\CMS\Toolbar\ToolbarHelper;
use Joomla\Input\Input;
use Joomla\Registry\Registry;

class PlgDPCalendarCSV extends SyncPlugin
{
	use ExportTrait;

	protected $identifier = 'c';

	/** @var CMSApplication */
	protected $app;

	public function onDPCalendarAfterToolbar(string $viewName, Toolbar $toolbar)
	{
		if (!in_array($viewName, ['events', 'bookings', 'tickets'])) {
			return;
		}
		ToolbarHelper::custom($viewName . '.csvexport', 'download', '', 'PLG_DPCALENDAR_CSV_EXPORT_CSV', false);
	}

	public function onDPCalendarBeforeExecute(Input $input)
	{
		$task = $input->get('task', '');
		if (strpos($task, 'csvexport') === false) {
			return;
		}

		list($action, $task) = explode('.', $task);

		$data = [];
		switch ($action) {
			case 'events':
				$data = $this->getEventData();
				break;
			case 'bookings':
				$data = $this->getBookingsData();
				break;
			case 'tickets':
				$data = $this->getTicketsData();
				break;
		}

		$f = fopen('php://memory', 'w');
		foreach ($data as $row) {
			fputcsv($f, $row, $this->params->get('export_separator', ','));
		}
		fseek($f, 0);

		header('Content-Type: application/csv');
		header('Content-Disposition: attachement; filename="' . $action . 's-' . date('YmdHi') . '.csv";');
		fpassthru($f);

		$this->app->close();
	}

	protected function getContent($calendarId, Date $startDate = null, Date $endDate = null, Registry $options)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return '';
		}

		/** @var Registry $params */
		$params = $calendar->params;

		try {
			// The content
			$content = $this->fetchContent(
				$params->get('uri', ''),
				[
					CURLOPT_SSL_VERIFYHOST => $params->get('ssl_verify') ? 2 : 0,
					CURLOPT_SSL_VERIFYPEER => $params->get('ssl_verify', 0)
				]
			);
			if ($content instanceof Exception) {
				$this->log($content->getMessage());

				return '';
			}

			// Remove UTF-8 BOM
			$content = str_replace("\xEF\xBB\xBF", '', $content);

			$lines = explode("\n", trim($content));
			if (!$lines) {
				return;
			}

			$k = [
				$params->get('all_day', 'all_day') => $params->get('all_day', 'all_day'),
				$params->get('start_date', 'start_date') => $params->get('start_date', 'start_date'),
				$params->get('end_date', 'end_date') => $params->get('end_date', 'end_date'),
				$params->get('show_end_time', 'show_end_time') => $params->get('show_end_time', 'show_end_time'),
				$params->get('title', 'title') => $params->get('title', 'title'),
				$params->get('description', 'description') => $params->get('description', 'description'),
				$params->get('rrule', 'rrule') => $params->get('rrule', 'rrule'),
				$params->get('location', 'location') => $params->get('location', 'location'),
				$params->get('alias', 'alias') => $params->get('alias', 'alias'),
				$params->get('color', 'color') => $params->get('color', 'color')
			];
			if ($params->get('has_header', true)) {
				$k = array_flip(str_getcsv(array_shift($lines), $params->get('delimiter', ','), $params->get('enclosure', '"')));
			}

			$text   = [];
			$text[] = 'BEGIN:VCALENDAR';
			foreach ($lines as $key => $line) {
				$line = trim($line);
				if (empty($line)) {
					continue;
				}
				try {
					$text[] = 'BEGIN:VEVENT';

					$data = str_getcsv($line, $params->get('delimiter', ','), $params->get('enclosure', '"'));

					$allDay = $this->getValue($calendar, $data, $k, 'all_day') == 1;

					$startDate = DPCalendarHelper::getDate($this->getValue($calendar, $data, $k, 'start_date'), $allDay);
					if ($allDay) {
						$text[] = 'DTSTART;VALUE=DATE:' . $startDate->format('Ymd');
					} else {
						$text[] = 'DTSTART:' . $startDate->format('Ymd\THis\Z');
					}
					$endDate = DPCalendarHelper::getDate($this->getValue($calendar, $data, $k, 'end_date'), $allDay);
					if ($allDay) {
						$text[] = 'DTEND;VALUE=DATE:' . $endDate->format('Ymd');
					} else {
						$text[] = 'DTEND:' . $endDate->format('Ymd\THis\Z');
					}

					$text[] = 'UID:' . md5($this->getValue($calendar, $data, $k, 'title') . 'CSV');
					$text[] = 'CATEGORIES:' . $this->params->get('title-' . $calendarId, 'Default');
					$text[] = 'SUMMARY:' . $this->getValue($calendar, $data, $k, 'title');
					$text[] = 'DESCRIPTION:' . str_replace("\n\r", '', $this->getValue($calendar, $data, $k, 'description'));

					$rrule = $this->getValue($calendar, $data, $k, 'rrule', false);
					if (empty($rrule)) {
						// Try out only rule for a more convenient way
						$old = $params->get('rrule');
						$params->set('rrule', 'rule');
						$rrule = $this->getValue($calendar, $data, $k, 'rrule', false);
						$params->set('rrule', $old);
					}
					if (!empty($rrule)) {
						$text[] = 'RRULE:' . $rrule;
					}

					$text[] = 'LOCATION:' . $this->getValue($calendar, $data, $k, 'location', false);

					$text[] = 'X-URL:' . $this->getValue($calendar, $data, $k, 'url', false);
					$text[] = 'X-ALIAS:' . $this->getValue($calendar, $data, $k, 'alias', false);
					$text[] = 'X-COLOR:' . $this->getValue($calendar, $data, $k, 'color', false);
					$text[] = 'X-SHOW-END-TIME:' . $this->getValue($calendar, $data, $k, 'show_end_time', false);

					$text[] = 'END:VEVENT';
				} catch (Exception $e) {
					$this->log($e->getMessage());
				}
			}
			$text[] = 'END:VCALENDAR';

			return $text;
		} catch (\Exception $e) {
			$this->log($e->getMessage());

			return '';
		}
	}

	private function getValue($calendar, $data, $k, $key, $throw = true)
	{
		$column = $calendar->params->get($key, $key);
		if (!key_exists($column, $k)) {
			if ($throw) {
				throw new InvalidArgumentException('Key ' . $key . ' not found to get the value from in the CSV file.');
			}

			return null;
		}
		if (!key_exists($k[$column], $data)) {
			if ($throw) {
				throw new InvalidArgumentException('Key ' . $key . ' not found to get the value from in the CSV file.');
			}

			return null;
		}

		return $data[$k[$column]];
	}
}
