<?php
/**
 * @package   DPCalendar
 * @copyright Copyright (C) 2022 Digital Peak GmbH. <https://www.digital-peak.com>
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 */

defined('_JEXEC') or die();

use DPCalendar\Helper\DPCalendarHelper;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Language;
use Joomla\CMS\Log\Log;
use Joomla\CMS\MVC\Model\BaseDatabaseModel;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\Table\Table;
use Joomla\CMS\User\User;
use Joomla\Component\Scheduler\Administrator\Event\ExecuteTaskEvent;
use Joomla\Component\Scheduler\Administrator\Task\Status;
use Joomla\Component\Scheduler\Administrator\Traits\TaskPluginTrait;
use Joomla\Event\SubscriberInterface;
use Joomla\Registry\Registry;
use Joomla\Utilities\ArrayHelper;

class PlgTaskDPCalendar extends CMSPlugin implements SubscriberInterface
{
	use TaskPluginTrait;

	protected const TASKS_MAP = [
		'plg_task_dpcalendar_ticket_send_certificates' => [
			'langConstPrefix' => 'PLG_TASK_DPCALENDAR_TASK_TICKET_SEND_CERTIFICATE',
			'form'            => 'ticket_send_certificates',
			'method'          => 'sendCertificates'
		],
		'plg_task_dpcalendar_booking_deletions' => [
			'langConstPrefix' => 'PLG_TASK_DPCALENDAR_TASK_BOOKING_DELETIONS',
			'form'            => 'booking_deletions',
			'method'          => 'deleteBookings'
		],
		'plg_task_dpcalendar_event_deletions' => [
			'langConstPrefix' => 'PLG_TASK_DPCALENDAR_TASK_EVENT_DELETIONS',
			'form'            => 'event_deletions',
			'method'          => 'deleteEvents',
		],
		'plg_task_dpcalendar_event_reminders' => [
			'langConstPrefix' => 'PLG_TASK_DPCALENDAR_TASK_EVENT_REMINDERS',
			'form'            => 'event_reminders',
			'method'          => 'sendEventReminders',
		]
	];

	public static function getSubscribedEvents(): array
	{
		return [
			'onTaskOptionsList'    => 'advertiseRoutines',
			'onExecuteTask'        => 'standardRoutineHandler',
			'onContentPrepareForm' => 'enhanceTaskItemForm',
		];
	}

	/** @var CMSApplication $app */
	protected $app;

	protected $autoloadLanguage = true;

	public function sendCertificates(ExecuteTaskEvent $event): int
	{
		if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
			return Status::INVALID_EXIT;
		}

		$params = new Registry($event->getArgument('params'));

		$this->app->getLanguage()->load('com_dpcalendar', JPATH_ADMINISTRATOR .  '/components/com_dpcalendar');
		Table::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/tables');
		BaseDatabaseModel::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/models');

		$db    = Factory::getDbo();
		$query = $db->getQuery(true)->select('t.id')->from('#__dpcalendar_tickets t');
		$query->innerJoin('#__dpcalendar_events e on t.event_id = e.id');

		$query->where('t.certificate_send_date is null');
		$query->where('t.state = 9');
		$query->where('e.end_date < ' . $db->quote(DPCalendarHelper::getDate()->toSql()));

		if ($calendars = $params->get('tickets_certificates_send_calendars', [])) {
			$query->where('e.catid in (' . implode(',', ArrayHelper::toInteger((array)$calendars)) . ')');
		}

		$db->setQuery($query);

		$result = $db->loadColumn();
		if ($result === false) {
			return Status::KNOCKOUT;
		}

		if (!$result) {
			return Status::OK;
		}

		$this->logTask('Found ' . count($result) . ' tickets to send certificate to', Log::DEBUG);

		foreach ($result as $id) {
			$model = BaseDatabaseModel::getInstance('Ticket', 'DPCalendarModel', ['ignore_request' => true]);
			$model->sendCertificate($model->getItem($id));
		}

		return Status::OK;
	}

	public function deleteBookings(ExecuteTaskEvent $event): int
	{
		if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
			return Status::INVALID_EXIT;
		}

		$this->app->getLanguage()->load('com_dpcalendar', JPATH_ADMINISTRATOR .  '/components/com_dpcalendar');
		Table::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/tables');
		BaseDatabaseModel::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/models');

		$params = new Registry($event->getArgument('params'));
		$date   = $params->get('deletions_date');
		$states = $params->get('deletions_states');

		if (!$date && !$states) {
			return Status::INVALID_EXIT;
		}

		$db    = Factory::getDbo();
		$query = $db->getQuery(true)->select('b.id')->from('#__dpcalendar_bookings b');

		if ($date) {
			$query->where('b.book_date < ' . $db->quote(DPCalendarHelper::getDate($date)->toSql()));
		}
		if ($states) {
			$query->where('b.state in (' . implode(',', ArrayHelper::toInteger((array)$states)) . ')');
		}
		$db->setQuery($query);

		$result = $db->loadColumn();

		if ($result === false) {
			return Status::KNOCKOUT;
		}

		if (!$result) {
			return Status::OK;
		}

		$this->logTask('Found ' . count($result) . ' bookings to delete', Log::DEBUG);

		BaseDatabaseModel::getInstance('Booking', 'DPCalendarModel', ['ignore_request' => true])->delete($result);

		return Status::OK;
	}

	public function deleteEvents(ExecuteTaskEvent $event): int
	{
		if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
			return Status::INVALID_EXIT;
		}

		$this->app->getLanguage()->load('com_dpcalendar', JPATH_ADMINISTRATOR .  '/components/com_dpcalendar');
		Table::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/tables');
		BaseDatabaseModel::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/models');

		$params    = new Registry($event->getArgument('params'));
		$startDate = $params->get('deletions_start_date');
		$endDate   = $params->get('deletions_end_date');
		$calendars = $params->get('deletions_calendars');
		$states    = $params->get('deletions_states');

		if (!$startDate && !$endDate && !$calendars && !$states) {
			return Status::INVALID_EXIT;
		}

		$db    = Factory::getDbo();
		$query = $db->getQuery(true)->select('e.id')->from('#__dpcalendar_events e');

		if ($startDate) {
			$query->where('e.start_date < ' . $db->quote(DPCalendarHelper::getDate($startDate)->toSql()));
		}
		if ($endDate) {
			$query->where('e.end_date < ' . $db->quote(DPCalendarHelper::getDate($endDate)->toSql()));
		}
		if ($states) {
			$query->where('e.state in (' . implode(',', ArrayHelper::toInteger((array)$states)) . ')');
		}
		if ($calendars) {
			$query->where('e.catid in (' . implode(',', ArrayHelper::toInteger((array)$calendars)) . ')');
		}
		$db->setQuery($query);

		$result = $db->loadColumn();

		if ($result === false) {
			return Status::KNOCKOUT;
		}

		if (!$result) {
			return Status::OK;
		}

		$this->logTask('Found ' . count($result) . ' events to delete', Log::DEBUG);

		BaseDatabaseModel::getInstance('AdminEvent', 'DPCalendarModel', ['ignore_request' => true])->publish($result, -2);
		BaseDatabaseModel::getInstance('AdminEvent', 'DPCalendarModel', ['ignore_request' => true])->delete($result);

		return Status::OK;
	}

	public function sendEventReminders(ExecuteTaskEvent $event): int
	{
		if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
			return Status::INVALID_EXIT;
		}

		$this->app->getLanguage()->load('com_dpcalendar', JPATH_ADMINISTRATOR .  '/components/com_dpcalendar');
		Table::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/tables');
		BaseDatabaseModel::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/models');
		BaseDatabaseModel::addIncludePath(JPATH_SITE . '/components/com_dpcalendar/models');

		$params = new Registry($event->getArgument('params'));

		$db    = Factory::getDbo();
		$query = $db->getQuery(true)->select('e.id')->from('#__dpcalendar_events e');
		$query->where('e.state = 1');

		$reminder = $params->get('reminder_time');
		if ($event->getArgument('subject')->get('last_execution') && (strpos($reminder, '+') === 0 || strpos($reminder, '-') === 0)) {
			$reminderDate = $db->quote(DPCalendarHelper::getDate($reminder)->format('Y-m-d H:i:00'));
			$lastRunDate  = DPCalendarHelper::getDate($event->getArgument('subject')->get('last_execution'));
			$lastRunDate->setTime($lastRunDate->format('H'), $lastRunDate->format('i'), 0, 0);
			$lastRunDate->modify($reminder)->format('Y-m-d H:i:00');
			$lastRunDate = $db->quote($lastRunDate);

			$query->where('(e.start_date > ' . $lastRunDate . ' and e.start_date <= ' . $reminderDate . ')');
		} else {
			$reminderDate = $db->quote(DPCalendarHelper::getDate($reminder)->format('Y-m-d H:i:00'));
			$query->where('e.start_date = ' . $reminderDate);
		}

		if ($calendars = $params->get('reminder_calendars', [])) {
			$query->where('e.catid in (' . implode(',', $calendars) . ')');
		}
		$db->setQuery($query);

		$result = $db->loadObjectList();

		$this->logTask('Found ' . count($result) . ' events to notify the author', Log::DEBUG);

		foreach ($result as $event) {
			$event = BaseDatabaseModel::getInstance('Event', 'DPCalendarModel')->getItem($event->id);
			if (!$event) {
				continue;
			}

			if ($params->get('send_author', 1)) {
				$params->set('subject_key', 'author_mail_subject');
				$params->set('subject_default', 'PLG_TASK_DPCALENDAR_TASK_EVENT_REMINDERS_AUTHOR_MAIL_SUBJECT_CONTENT');
				$params->set('message_key', 'author_mail_message');
				$params->set('message_default', 'PLG_TASK_DPCALENDAR_TASK_EVENT_REMINDERS_AUTHOR_MAIL_MESSAGE_CONTENT');
				$user = Factory::getUser($event->created_by);
				$this->send($event, $user->email, $user, $params);
			}

			if ($params->get('send_attendee', 1)) {
				$params->set('subject_key', 'attendee_mail_subject');
				$params->set('subject_default', 'PLG_TASK_DPCALENDAR_TASK_EVENT_REMINDERS_ATTENDEE_MAIL_SUBJECT_CONTENT');
				$params->set('message_key', 'attendee_mail_message');
				$params->set('message_default', 'PLG_TASK_DPCALENDAR_TASK_EVENT_REMINDERS_ATTENDEE_MAIL_MESSAGE_CONTENT');
				foreach ($event->tickets as $ticket) {
					$this->send($event, $ticket->email, Factory::getUser($ticket->user_id), $params);
				}
			}
		}

		return Status::OK;
	}

	private function send(stdClass $event, string $email, User $user, Registry $params)
	{
		try {
			$this->logTask('Starting to send out the notificaton for the event with the id: ' . $event->id, Log::DEBUG);

			$this->logTask('Settig up the texts', Log::DEBUG);

			$siteLanguage = ComponentHelper::getParams('com_languages')->get('site', $this->app->get('language', 'en-GB'));
			$lang         = Language::getInstance($user->getParam('language', $siteLanguage), $this->app->get('debug_lang'));
			$this->app->loadLanguage($lang);
			$lang->load('plg_task_dpcalendar', JPATH_PLUGINS . '/task/dpcalendar');

			$subject = DPCalendarHelper::renderEvents(
				[$event],
				DPCalendarHelper::getStringFromParams(
					$params->get('subject_key'),
					$params->get('subject_default'),
					$params,
					$this->app->getLanguage()
				)
			);

			$body = DPCalendarHelper::renderEvents(
				[$event],
				DPCalendarHelper::getStringFromParams(
					$params->get('message_key'),
					$params->get('message_default'),
					$params,
					$this->app->getLanguage()
				),
				null,
				['sitename' => $this->app->get('sitename')]
			);

			$this->logTask('Sending the mail to ' . $email, Log::DEBUG);

			$mailer = clone Factory::getMailer();
			$mailer->setSubject($subject);
			$mailer->setBody($body);
			$mailer->IsHTML(true);
			$mailer->AddAddress($email);
			$mailer->Send();
		} catch (Exception $e) {
			$this->logTask('Error sending mail! Exception is: ' . PHP_EOL . $e, Log::ERROR);
		}

		$this->logTask('Finished to send out the notificaton for the event with the id: ' . $event->id, Log::DEBUG);
	}
}
