<?php
/**
 * @package   DPCalendar
 * @copyright Copyright (C) 2022 Digital Peak GmbH. <https://www.digital-peak.com>
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 */

defined('_JEXEC') or die;

use DPCalendar\Helper\DPCalendarHelper;
use Joomla\CMS\MVC\Model\BaseDatabaseModel;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\Table\Table;

define('RSFORM_FIELD_DPCALENDAR_RSFORM', 77832);

class PlgSystemDPCalendarRSForm extends CMSPlugin
{
	protected $app;
	protected $autoloadLanguage = true;

	public function __construct($subject, array $config = [])
	{
		parent::__construct($subject, $config);

		// Ensure RSForms is correctly loaded
		if (!class_exists('RSFormProHelper')) {
			if (!file_exists(JPATH_ADMINISTRATOR . '/components/com_rsform/helpers/rsform.php')) {
				return;
			}

			require_once JPATH_ADMINISTRATOR . '/components/com_rsform/helpers/rsform.php';
		}
	}

	public function onRsformBackendAfterCreateFieldGroups(&$fieldGroups, $self)
	{
		// Add the entry in the sidebar to create the component
		$fieldGroups['advanced']->fields[] = (object) [
			'id'   => RSFORM_FIELD_DPCALENDAR_RSFORM,
			'name' => $this->app->getLanguage()->_('PLG_SYSTEM_DPCALENDARRSFORM_LABEL'),
			'icon' => 'rsficon rsficon-calendar1'
		];
	}

	public function onRsformFrontendAfterStoreSubmissions($data)
	{
		// Check if DPCalendar is available
		if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
			return;
		}

		// Setup RSForms
		require_once JPATH_ADMINISTRATOR . '/components/com_rsform/helpers/submissions.php';

		// Get the data
		$submission     = RSFormProSubmissionsHelper::getSubmission($data['SubmissionId']);
		$components     = RSFormProHelper::getComponents($data['formId']);
		$componentProps = RSFormProHelper::getComponentProperties($components);
		$form           = RSFormProHelper::getForm($data['formId']);

		// Setup DPCalendar
		Table::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/tables', 'DPCalendarTable');
		BaseDatabaseModel::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/models', 'DPCalendarModel');

		// Loop over the components
		foreach ($components as $component) {
			// Accept only the DPCalendar component
			if ($component->ComponentTypeName !== 'dpcalendarreservation') {
				continue;
			}

			// Check if there is a value
			if (!array_key_exists($component->name, $submission->values)) {
				continue;
			}

			// Check if there are properties for the component
			if (!array_key_exists($component->ComponentId, $componentProps)) {
				continue;
			}

			// Init fields
			$title       = $form->FormId . ' ' . $form->FormName;
			$description = '';

			// Loop over the values from the submission to compile the description and set the title
			foreach ($submission->values as $name => $value) {
				// Ignore own values from own component
				if ($name === $component->name || $name === $component->name . '-date') {
					continue;
				}

				// Check if data is valid
				if ($name === 'formId' || empty($value)) {
					continue;
				}

				// Ignore the title field
				if ($name === $componentProps[$component->ComponentId]['EVENT_TITLE_FIELD_NAME']) {
					$title = $value;
					continue;
				}

				// Add the value to the description
				$description .= '<p>';
				$description .= $this->app->getLanguage()->_($name) . ': ';
				$description .= $this->app->getLanguage()->_($value);
				$description .= '</p>';
			}

			// The timezone for the dates
			$timezone = DPCalendarHelper::getDate()->getTimezone()->getName();

			// Extract start and end time
			list($start, $end) = explode('-', $submission->values[$component->name]);

			// Compile the dates
			$start = DPCalendarHelper::getDate($submission->values[$component->name . '-date'] . ' ' . $start, false, $timezone);
			$end   = DPCalendarHelper::getDate($submission->values[$component->name . '-date'] . ' ' . $end, false, $timezone);

			// Load the model and save the data
			$model = BaseDatabaseModel::getInstance('Adminevent', 'DPCalendarModel', ['ignore_request' => true]);
			$model->save([
				'title'          => $title,
				'start_date'     => $start->toSql(),
				'end_date'       => $end->toSql(),
				'catid'          => $componentProps[$component->ComponentId]['CALENDAR_ID'],
				'access'         => $componentProps[$component->ComponentId]['EVENT_ACCESS'] ?? 1,
				'access_content' => $componentProps[$component->ComponentId]['EVENT_ACCESS_CONTENT'] ?? 1,
				'all_day'        => 0,
				'description'    => $description,
				'capacity'       => 0,
				// When there are payments involved, then the event needs to be unpublished
				'state'      => defined('RSFORM_FIELD_PAYMENT_TOTAL') && RSFormProHelper::componentExists($form->FormId, RSFORM_FIELD_PAYMENT_TOTAL) ? 0 : 1,
				'xreference' => 'rsform-' . $submission->SubmissionId . '-' . $component->ComponentId
			]);
		}
	}

	public function onRsformAfterConfirmPayment($submissionId)
	{
		// Check if DPCalendar is available
		if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
			return [];
		}

		Table::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/tables', 'DPCalendarTable');

		// Publish the event when payment is confirmed
		foreach ($this->getEvents($submissionId) as $event) {
			Table::getInstance('Event', 'DPCalendarTable')->publish([$event->id], 1);
		}
	}

	public function onRsformDoPayment($value, $formId, $submissionId, $price)
	{
		// Check if DPCalendar is available
		if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
			return [];
		}

		Table::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/tables', 'DPCalendarTable');

		// Unpublish the event when payment process starts
		foreach ($this->getEvents($submissionId) as $event) {
			Table::getInstance('Event', 'DPCalendarTable')->publish([$event->id], 0);
		}
	}

	public function onRsformDeleteSubmissionFiles($fields, $formId, $submissionIds)
	{
		// Check if DPCalendar is available
		if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
			return [];
		}

		Table::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/tables', 'DPCalendarTable');

		// Delete events of the deleted submissions
		foreach ($submissionIds as $id) {
			foreach ($this->getEvents($id) as $event) {
				Table::getInstance('Event', 'DPCalendarTable')->delete($event->id);
			}
		}
	}

	public function onRsformAfterCreatePlaceholders($data)
	{
		// Check if DPCalendar is available
		if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
			return [];
		}

		Table::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/tables', 'DPCalendarTable');

		$dateFormat = DPCalendarHelper::getComponentParameter('event_date_format', 'd.m.Y');
		$dateFormat .= ' ' . DPCalendarHelper::getComponentParameter('event_time_format', 'H:i');

		$components = RSFormProHelper::getComponents($data['form']->FormId);
		foreach ($components as $component) {
			// Accept only the DPCalendar component
			if ($component->ComponentTypeName !== 'dpcalendarreservation') {
				continue;
			}

			$event = Table::getInstance('Event', 'DPCalendarTable');
			$event->load(['xreference' => 'rsform-' . $data['submission']->SubmissionId . '-' . $component->ComponentId]);
			if (!$event->id) {
				continue;
			}

			$data['placeholders'][] = '{' . $component->name . ':start_date}';
			$data['values'][]       = $event->start_date;

			$data['placeholders'][] = '{' . $component->name . ':end_date}';
			$data['values'][]       = $event->end_date;

			$data['placeholders'][] = '{' . $component->name . ':start_date_formatted}';
			$data['values'][]       = DPCalendarHelper::getDate($event->start_date)->format($dateFormat, true);

			$data['placeholders'][] = '{' . $component->name . ':end_date_formatted}';
			$data['values'][]       = DPCalendarHelper::getDate($event->end_date)->format($dateFormat, true);

			$data['placeholders'][] = '{' . $component->name . ':date_range}' ;
			$data['values'][]       = str_replace(PHP_EOL, '', trim(strip_tags(DPCalendarHelper::getDateStringFromEvent($event))));
		}
	}

	/**
	 * Returns events for the given submission.
	 *
	 * @param string $submissionId
	 *
	 * @return array
	 */
	private function getEvents(string $submissionId): array
	{
		// Check if DPCalendar is available
		if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
			return [];
		}

		// Setup DPCalendar
		BaseDatabaseModel::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/models', 'DPCalendarModel');

		$model = BaseDatabaseModel::getInstance('Adminevents', 'DPCalendarModel', ['ignore_request' => true]);
		$model->setState('filter.xreference', 'rsform-' . $submissionId . '%');

		return $model->getItems();
	}
}
