<?php
/**
 * @package   DPCalendar
 * @copyright Copyright (C) 2014 Digital Peak GmbH. <https://www.digital-peak.com>
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 */
defined('_JEXEC') or die();

if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
	return;
}

use DigitalPeak\ThinHTTP as HTTP;
use Joomla\CMS\Date\Date;
use Joomla\Registry\Registry;

class PlgDPCalendarFacebook extends \DPCalendar\Plugin\SyncPlugin
{
	protected $identifier = 'fb';

	/** @var \Joomla\CMS\Application\CMSApplication */
	protected $app;

	private $fbUrl = 'https://graph.facebook.com/v8.0/';

	protected function getSyncToken($calendar)
	{
		// We are doing a raw request because FB supports etags in their
		// response headers https://developers.facebook.com/blog/post/627
		$http     = new HTTP();
		$response = $http->get(
			$this->fbUrl . $calendar->params->get('page-id') . '/events',
			$calendar->params->get('access-token')
		);
		if (!empty($response->error)) {
			throw new Exception($response->error->message);
		}

		$syncToken = null;
		if (key_exists('etag', $response->dp->headers) && !empty($response->dp->headers['etag'])) {
			$syncToken = $response->dp->headers->etag[0];
		} else if (key_exists('last-modified', $response->dp->headers) && !empty($response->dp->headers['last-modified'])) {
			$syncToken = $response->dp->headers['last-modified'][0];
		}

		if ($syncToken) {
			return $syncToken;
		}

		return parent::getSyncToken($calendar);
	}

	/*
	 * Because the end date of Facebook is exclusive we need to
	 * overwrite that function.
	 */
	public function fetchEvent($eventId, $calendarId)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return '';
		}

		$pos = strrpos($eventId, '_');
		if ($pos === false) {
			return null;
		}

		try {
			$url     = $this->fbUrl . trim(substr($eventId, 0, $pos));
			$url     .= '?fields=name,cover,id,start_time,end_time,timezone,place,description,updated_time';
			$fbEvent = (new HTTP())->get($url, $calendar->params->get('access-token'));
			if (!empty($fbEvent->error)) {
				throw new Exception($fbEvent->error->message);
			}

			$event      = $this->createEvent($fbEvent->id, $calendarId);
			$event->uid = $fbEvent->id;

			$event->all_day = strlen($fbEvent->start_time) == 10;
			$timezone       = 'UTC';
			if (!empty($fbEvent->timezone)) {
				$timezone = $fbEvent->timezone;
			}

			$startDate         = DPCalendarHelper::getDate($fbEvent->start_time, $event->all_day, $timezone);
			$event->start_date = $startDate->toSql();

			$endDate = null;
			if (empty($fbEvent->end_time)) {
				$endDate = clone $startDate;
				if (!$event->all_day) {
					$endDate->setTime(23, 59, 59);
				}
			} else {
				$endDate = DPCalendarHelper::getDate($fbEvent->end_time, $event->all_day, $timezone);
			}
			$event->end_date = $endDate->toSql();

			$event->title = $fbEvent->name;
			if (!empty($fbEvent->description)) {
				$event->description = DPCalendarHelper::parseHtml($this->replaceNl(nl2br($fbEvent->description)));
			}

			// Using place field for location information
			if (!empty($fbEvent->place) && !empty($fbEvent->place->location)) {
				$location = \DPCalendar\Helper\Location::format($fbEvent->place->location);
				$title    = $location;
				if (!empty($fbEvent->place->location->latitude) && !empty($fbEvent->place->location->longitude)) {
					$location = $fbEvent->place->location->latitude . ',' . $fbEvent->place->location->longitude;
				}
				$locationObject   = \DPCalendar\Helper\Location::get($location, true, $title);
				$event->locations = [$locationObject];
			}

			$event->url = 'https://www.facebook.com/events/' . $fbEvent->id;

			if (!empty($fbEvent->cover)) {
				$event->images                     = new stdClass();
				$event->images->image_full_alt     = '';
				$event->images->image_full_caption = '';
				$event->images->image_full         = $fbEvent->cover->source;
			}

			return $event;
		} catch (Exception $e) {
			$this->log($e->getMessage());
		}

		return null;
	}

	public function fetchEvents($calendarId, Date $startDate = null, Date $endDate = null, Registry $options)
	{
		$events = parent::fetchEvents($calendarId, $startDate, $endDate, $options);
		foreach ($events as $event) {
			$event->description = DPCalendarHelper::parseHtml($event->description);
		}

		return $events;
	}

	protected function getContent($calendarId, Date $startDate = null, Date $endDate = null, Registry $options)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return '';
		}

		$url = $this->fbUrl . trim($calendar->params->get('page-id'));
		$url .= '/events?fields=name,cover,id,start_time,end_time,timezone,place,description,updated_time&limit=10000';
		if ($startDate == null) {
			$url .= '&since=' . urlencode(DPCalendarHelper::getDate()->format('c'));
		} else {
			// Since is exclusive so we need to subtract a second
			$startDate->modify('-1 second');
			$url .= '&since=' . urlencode($startDate->format('c'));
		}

		if ($endDate !== null) {
			$url .= '&until=' . urlencode($endDate->format('c'));
		}

		$events = (new HTTP())->get($url, $calendar->params->get('access-token'));
		if (!empty($events->error)) {
			throw new Exception($events->error->message);
		}

		$text   = [];
		$text[] = 'BEGIN:VCALENDAR';
		foreach ($events->data as $event) {
			$text[] = 'BEGIN:VEVENT';

			$allDay   = strlen($event->start_time) == 10;
			$timezone = 'UTC';
			if (!empty($event->timezone)) {
				$timezone = $event->timezone;
			}

			$startDate = DPCalendarHelper::getDate($event->start_time, $allDay, $allDay ? null : $timezone);
			if ($allDay) {
				$text[] = 'DTSTART;VALUE=DATE:' . $startDate->format('Ymd');
			} else {
				$text[] = 'DTSTART:' . $startDate->format('Ymd\THis\Z');
			}

			$endDate = null;
			if (empty($event->end_time)) {
				$endDate = clone $startDate;
				if ($allDay) {
					$endDate->modify('+1 day');
				} else {
					$endDate->setTime(23, 59, 59);
				}
			} else {
				$endDate = DPCalendarHelper::getDate($event->end_time, $allDay, $allDay ? null : $timezone);
			}
			if ($allDay) {
				$text[] = 'DTEND;VALUE=DATE:' . $endDate->format('Ymd');
			} else {
				$text[] = 'DTEND:' . $endDate->format('Ymd\THis\Z');
			}

			$text[] = 'UID:' . $event->id;
			$text[] = 'CATEGORIES:' . $calendar->params->get('title', 'Default');
			$text[] = 'SUMMARY:' . $event->name;
			if (!empty($event->description)) {
				$text[] = 'DESCRIPTION:' . $this->replaceNl(nl2br($event->description));
			}

			$text[] = 'X-URL:https://www.facebook.com/events/' . $event->id;

			// Using place field for location information
			if (!empty($event->place) && !empty($event->place->location)) {
				$location = $event->place->location;
				$text[]   = 'LOCATION:' . \DPCalendar\Helper\Location::format(json_decode(json_encode($location), false));

				if (!empty($location->latitude) && !empty($location->longitude)) {
					$text[] = 'GEO:' . $location->latitude . ';' . $location->longitude;
				}
			}

			if (!empty($event->cover)) {
				$text[] = 'X-IMAGE:' . $event->cover->source;
			}

			$text[] = 'END:VEVENT';
		}
		$text[] = 'END:VCALENDAR';

		return $text;
	}

	public function onContentBeforeDisplay($context, $event)
	{
		if ($context != 'com_dpcalendar.event') {
			return;
		}

		$doc = $this->app->getDocument();
		$doc->addCustomTag('<meta property="og:type" content="event"/>');
		$doc->addCustomTag('<meta property="og:title" content="' . htmlspecialchars($event->title) . '"/>');
		$doc->addCustomTag('<meta property="og:description" content="' . strip_tags($event->description) . '"/>');
		$doc->addCustomTag('<meta property="og:url" content="' . str_replace('"', '&quot;', JURI::current()) . '"/>');
		$doc->addCustomTag('<meta property="event:start_time" content="' . DPCalendarHelper::getDate($event->start_date)->format('Y-m-d\TH:i:s\Z') . '"/>');

		if (!empty($event->images->image_full)) {
			$doc->addCustomTag('<meta property="og:image" content="' . htmlspecialchars(JURI::root() . $event->images->image_full) . '"/>');
		}
		if (!empty($event->images->image_full_alt)) {
			$doc->addCustomTag('<meta property="og:image:alt" content="' . htmlspecialchars($event->images->image_full_alt) . '"/>');
		}
		if (!empty($event->images->image_full_width)) {
			$doc->addCustomTag('<meta property="og:image:width" content="' . (int)$event->images->image_full_width . '"/>');
		}
		if (!empty($event->images->image_full_height)) {
			$doc->addCustomTag('<meta property="og:image:height" content="' . (int)$event->images->image_full_height . '"/>');
		}

		foreach ($this->fetchCalendars() as $calendar) {
			$doc->addCustomTag('<meta property="fb:app_id" content="' . $calendar->params->get('app-id') . '"/>');
			break;
		}
	}

	public function onContentDisplayEventList($context, $calendarIds, $events)
	{
		if ($context != 'com_dpcalendar.calendar' || !$calendarIds) {
			return;
		}

		$doc = $this->app->getDocument();
		foreach ($calendarIds as $calendarId) {
			$calendar = \DPCalendar\Helper\DPCalendarHelper::getCalendar($calendarId);
			$doc->addCustomTag('<meta property="og:type" content="article"/>');
			$doc->addCustomTag('<meta property="og:title" content="' . htmlspecialchars($calendar->title) . '"/>');
			$doc->addCustomTag('<meta property="og:description" content="' . strip_tags($calendar->description) . '"/>');
			$doc->addCustomTag('<meta property="og:url" content="' . str_replace('"', '&quot;', JURI::current()) . '"/>');

			if (!empty($calendar->params->get('image'))) {
				$doc->addCustomTag('<meta property="og:image" content="' . htmlspecialchars(JURI::root() . $calendar->params->get('image')) . '"/>');
			}
			if (!empty($calendar->params->get('image_alt'))) {
				$doc->addCustomTag('<meta property="og:image:alt" content="' . htmlspecialchars($calendar->params->get('image_alt')) . '"/>');
			}

			foreach ($this->fetchCalendars() as $calendar) {
				$doc->addCustomTag('<meta property="fb:app_id" content="' . $calendar->params->get('app-id') . '"/>');
			}

			if (!empty($calendar->params->get('image')) || !$events) {
				break;
			}

			$event = reset($events);
			if (!empty($event->images->image_full)) {
				$doc->addCustomTag('<meta property="og:image" content="' . htmlspecialchars(JURI::root() . $event->images->image_full) . '"/>');
			}
			if (!empty($event->images->image_full_width)) {
				$doc->addCustomTag('<meta property="og:image:width" content="' . (int)$event->images->image_full_width . '"/>');
			}
			if (!empty($event->images->image_full_height)) {
				$doc->addCustomTag('<meta property="og:image:height" content="' . (int)$event->images->image_full_height . '"/>');
			}
			if (!empty($event->images->image_full_alt)) {
				$doc->addCustomTag('<meta property="og:image:alt" content="' . htmlspecialchars($event->images->image_full_alt) . '"/>');
			}
			break;
		}

		foreach ($this->fetchCalendars() as $calendar) {
			$doc->addCustomTag('<meta property="fb:app_id" content="' . $calendar->params->get('app-id') . '"/>');
			break;
		}
	}

	public function onContentAfterSave($context, $item, $isNew)
	{
		if ($context != 'com_dpcalendar.event') {
			return true;
		}

		foreach ($this->fetchCalendars() as $calendar) {
			if ($isNew && $calendar->params->get('creationeventupdate', 1)) {
				$this->publishEvent('create', $item, $calendar);
			}

			if (!$isNew && $calendar->params->get('updateeventupdate', 1)) {
				$this->publishEvent('save', $item, $calendar);
			}
		}
	}

	public function onContentAfterDelete($context, $item)
	{
		if ($context != 'com_dpcalendar.event') {
			return true;
		}

		foreach ($this->fetchCalendars() as $calendar) {
			if ($calendar->params->get('deletioneventupdate', 0)) {
				$this->publishEvent('delete', $item, $calendar);
			}
		}
	}

	private function publishEvent($action, $event, $calendar)
	{
		// We don't publish external events
		if (!is_numeric($event->catid)) {
			return;
		}

		$params = $calendar->params;
		$ids    = $params->get('ids');
		if (!$ids) {
			return;
		}

		if ($params->get('eventsinsubcat', 0)) {
			$sids = $ids;
			foreach ($sids as $id) {
				$cat = JCategories::getInstance('DPCalendar')->get($id);
				if (!$cat) {
					continue;
				}

				foreach ($cat->getChildren() as $ch) {
					$ids[] = $ch->id;
				}
			}
		}

		if (!in_array($event->catid, $ids)) {
			return;
		}

		$status   = '';
		$eventurl = DPCalendarHelperRoute::getEventRoute($event->id, $event->catid, true);
		$eventurl = str_replace('administrator/', '', $eventurl);
		switch ($action) {
			case 'create':
				$status = JText::sprintf('PLG_DPCALENDAR_FACEBOOK_STATUS_EVENT_CREATED', $event->title, $eventurl);
				break;
			case 'save':
				$status = JText::sprintf('PLG_DPCALENDAR_FACEBOOK_STATUS_EVENT_UPDATED', $event->title, $eventurl);
				break;
			case 'delete':
				$status = JText::sprintf('PLG_DPCALENDAR_FACEBOOK_STATUS_EVENT_DELETED', $event->title);
				break;
		}

		try {
			$reply = (new HTTP())->post(
				$this->fbUrl . trim($calendar->params->get('page-id')) . '/feed',
				http_build_query(['message' => $status]),
				$calendar->params->get('access-token')
			);
			if (!empty($reply->error)) {
				throw new Exception($reply->error->message);
			}

			$publishUrl = '';
			$parts      = explode('_', $reply->id);
			if (count($parts) >= 2) {
				$publishUrl = 'https://www.facebook.com/permalink.php?story_fbid=' . $parts[1] . '&id=' . $parts[0];
			}
			$this->app->enqueueMessage(
				JText::sprintf('PLG_DPCALENDAR_FACEBOOK_PUBLISH_SUCCESS', $publishUrl, $calendar->title)
			);
		} catch (Exception $e) {
			$this->log(JText::sprintf('PLG_DPCALENDAR_FACEBOOK_PUBLISH_ERROR', $e->getMessage()));
		}
	}

	public function import()
	{
		$session = JFactory::getSession(['expire' => 30]);

		// If we are on the callback from facebook don't save
		if (!$this->app->input->get('code')) {
			$params = $this->app->input->get('params', ['app-id' => null, 'secret' => null], 'array');
			$session->set('app-id', $params['app-id'], $this->_name);
			$session->set('secret', $params['secret'], $this->_name);
			$session->set('page-id', $params['page-id'], $this->_name);
		}
		$appId  = $session->get('app-id', null, $this->_name);
		$secret = $session->get('secret', null, $this->_name);
		$pageId = $session->get('page-id', null, $this->_name);

		if ($this->app->input->get('code')) {
			$session->set('app-id', null, $this->_name);
			$session->set('secret', null, $this->_name);
			$session->set('page-id', null, $this->_name);
		}

		try {
			$uri         = JUri::getInstance();
			$redirectUrl = urlencode($uri->toString(['scheme', 'host', 'port', 'path'])
				. '?option=com_dpcalendar&task=plugin.action&dpplugin=facebook&action=import');
			if (!$this->app->input->get('code')) {
				$loginUrl = 'https://www.facebook.com/v8.0/dialog/oauth?client_id=' . $appId;
				$loginUrl .= '&redirect_uri=' . $redirectUrl;
				$loginUrl .= '&state=' . $session->getToken();
				$loginUrl .= '&scope=email,pages_read_user_content,pages_manage_posts';
				$this->app->redirect($loginUrl);
				$this->app->close();

				return;
			}

			$http = new HTTP();

			$url   = $this->fbUrl . 'oauth/access_token?client_id=' . $appId;
			$url   .= '&redirect_uri=' . $redirectUrl;
			$url   .= '&client_secret=' . $secret;
			$url   .= '&code=' . $this->app->input->get('code');
			$token = $http->get($url);
			if (!empty($token->error)) {
				throw new Exception($token->error->message);
			}

			$user = $http->get($this->fbUrl . 'me', $token->access_token);
			if (!empty($user->error)) {
				throw new Exception($user->error->message);
			}

			$pages = $http->get($this->fbUrl . $user->id . '/accounts', $token->access_token);
			if (!empty($pages->error)) {
				throw new Exception($pages->error->message);
			}

			$pages = $pages->data;
			if ($pageId) {
				if (!is_numeric($pageId)) {
					$pageId = trim(substr($pageId, strrpos($pageId, '/')), '/');
				}
				$data = $http->get($this->fbUrl . $pageId, $token->access_token);
				if (!empty($data->id)) {
					$data->access_token = $token->access_token;
					$pages              = [$data];
				} else {
					$this->app->enqueueMessage(JText::printf('PLG_DPCALENDAR_FACEBOOK_MESSAGE_NO_PAGE_ID', $pageId), 'warning');
				}
			}

			foreach ($pages as $pageData) {
				$page = $http->get(
					$this->fbUrl . $pageData->id . '?fields=id,name,access_token,about,description,description_html',
					$token->access_token
				);
				if (!empty($page->error)) {
					throw new Exception($page->error->message);
				}

				$model  = JModelLegacy::getInstance('Extcalendar', 'DPCalendarModel');
				$params = new Registry();
				$params->set('app-id', $appId);
				$params->set('secret', $secret);
				$params->set('page-id', $page->id);
				$params->set('access-token', $page->access_token ?: $token->access_token);

				$table = JTable::getInstance('Extcalendar', 'DPCalendarTable');
				$table->load(['title' => $page->name]);

				$data = [];
				if ($table->id > 0) {
					$data['id'] = $table->id;
				}
				$data['title'] = $page->name;
				if (!empty($page->description_html)) {
					$data['description'] = $page->description_html;
				} else if (!empty($page->description)) {
					$data['description'] = $page->description;
				} else if (!empty($page->about)) {
					$data['description'] = $page->about;
				}
				$data['color']  = substr(dechex(crc32($data['title'])), 0, 6);
				$data['params'] = $params->toString();
				$data['plugin'] = 'facebook';

				if (!$model->save($data)) {
					$this->app->enqueueMessage($model->getError(), 'warning');
				}
			}

			$this->app->redirect(
				JFactory::getSession()->get(
					'extcalendarOrigin',
					'index.php?option=com_dpcalendar&view=extcalendars&dpplugin=facebook&tmpl=' . $this->app->input->get('tmpl'),
					'DPCalendar'
				)
			);
			$this->app->close();
		} catch (Exception $e) {
			$this->app->enqueueMessage($e->getMessage(), 'warning');
			$this->app->redirect(
				JFactory::getSession()->get(
					'extcalendarOrigin',
					'index.php?option=com_dpcalendar&view=extcalendars&dpplugin=facebook&tmpl=' . $this->app->input->get('tmpl'),
					'DPCalendar'
				)
			);
		}
	}
}
