<?php
/**
 * @package   DPCalendar
 * @copyright Copyright (C) 2014 Digital Peak GmbH. <https://www.digital-peak.com>
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 */
defined('_JEXEC') or die();

if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
	return;
}

use DigitalPeak\ThinHTTP as HTTP;
use DPCalendar\Helper\Location;
use DPCalendar\Plugin\SyncPlugin;
use Joomla\CMS\Date\Date;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Model\BaseDatabaseModel;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\Uri\Uri;
use Joomla\Registry\Registry;

class PlgDPCalendarGoogle extends SyncPlugin
{
	protected $identifier = 'g';

	/** @var \Joomla\CMS\Application\CMSApplication */
	protected $app;

	protected function getSyncToken($calendar)
	{
		$params               = [];
		$params['updatedMin'] = DPCalendarHelper::getDate($calendar->sync_date)->format('c');
		$params['maxResults'] = 1;
		$obj                  = $this->callAPI($calendar, '/calendars/' . urlencode($calendar->params->get('calendarId')) . '/events', $params);

		return $obj->etag;
	}

	public function fetchEvent($eventId, $calendarId)
	{
		return $this->fetchEventFromGoogle(false, $eventId, $calendarId);
	}

	private function fetchEventFromGoogle($returnGoogleEvent, $eventId, $calendarId)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return null;
		}

		$params = [];

		try {
			$eventId = urldecode($eventId);
			$pos     = strrpos($eventId, '_');
			if ($pos === false) {
				return null;
			}

			$googleEvent = $this->callAPI(
				$calendar,
				'/calendars/' . urlencode($calendar->params->get('calendarId')) . '/events/' . substr($eventId, 0, $pos)
			);
			if (!$googleEvent) {
				return null;
			}
			$obj = $googleEvent;

			// Check if we need an instance
			$s = substr($eventId, $pos + 1);
			if ($s != '0' && !empty($googleEvent->recurrence)) {
				$startDate = null;
				if (strlen($s) == 8) {
					$startDate = Factory::getDate(substr($s, 0, 4) . '-' . substr($s, 4, 2) . '-' . substr($s, 6, 2) . ' 00:00');
				} else {
					$startDate = Factory::getDate(
						substr($s, 0, 4) . '-' . substr($s, 4, 2) . '-' . substr($s, 6, 2) . ' ' . substr($s, 8, 2) . ':' . substr($s, 10, 2)
					);
				}

				// Exact date google has problems on recurring events
				$startDate->modify('-1 second');
				$params['timeMin'] = $startDate->format('c');
				$startDate->modify('+1 ' . (strlen($s) == 8 ? 'day' : 'hour'));
				$params['timeMax']    = $startDate->format('c');
				$params['maxResults'] = 1;

				$obj = $this->callAPI(
					$calendar,
					'/calendars/' . urlencode($calendar->params->get('calendarId')) . '/events/' . substr($eventId, 0, $pos) . '/instances',
					$params
				);
				if ($obj && count($obj->items) > 0) {
					$googleEvent = $obj->items[0];
				} else {
					$googleEvent = null;
				}
			}

			if (!$googleEvent) {
				return null;
			}

			if ($returnGoogleEvent) {
				return $googleEvent;
			}

			$colors = $this->callAPI($calendar, '/colors');
			if (!empty($colors->event)) {
				$colors = $colors->event;
			}

			return $this->createEventFromGoogle($googleEvent, $calendar, $obj, $colors);
		} catch (Exception $e) {
			$this->log($e);
		}
	}

	public function fetchEvents($calendarId, Date $startDate = null, Date $endDate = null, Registry $options)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return [];
		}

		if ($startDate == null) {
			$startDate = DPCalendarHelper::getDate();
		}

		$params = [];
		if ($startDate != null) {
			// Exact date google has problems on recurring events
			$startDate->modify('-1 second');
			$params['timeMin'] = $startDate->format('c');
		}
		if ($endDate != null) {
			$params['timeMax'] = $endDate->format('c');
		}
		if ($options->get('filter')) {
			$params['q'] = $options->get('filter');
		}
		$params['maxResults'] = $options->get('limit', 1000);

		if ($options->get('expand', true)) {
			$params['singleEvents'] = 'true';
			$params['orderBy']      = 'startTime';
		} else {
			$params['showDeleted'] = 'true';
		}

		try {
			$obj = $this->callAPI($calendar, '/calendars/' . urlencode($calendar->params->get('calendarId')) . '/events', $params);
			if (empty($obj->items)) {
				return [];
			}
			$googleEvents = $obj->items;

			$order = strtolower($options->get('order', 'asc'));
			usort(
				$googleEvents,
				function ($event1, $event2) use ($order) {
					if (!$event1->start || !$event2->start) {
						return 0;
					}

					$first  = $event1;
					$second = $event2;
					if (strtolower($order) == 'desc') {
						$first  = $event2;
						$second = $event1;
					}

					return strcmp(
						!empty($first->start->date) ? $first->start->date : $first->start->dateTime,
						!empty($second->start->date) ? $second->start->date : $second->start->dateTime
					);
				}
			);

			$colors = $this->callAPI($calendar, '/colors');
			if (!empty($colors->event)) {
				$colors = $colors->event;
			}
			$events = [];
			foreach ($googleEvents as $googleEvent) {
				//Check if it is a normal deleted event
				if ($googleEvent->status == 'cancelled' && empty($googleEvent->recurringEventId)) {
					// Ignore it, will be sent when expand is false
					continue;
				}

				$event = $this->createEventFromGoogle($googleEvent, $calendar, $obj, $colors);
				if (!$event || !$this->matchLocationFilterEvent($event, $options)) {
					continue;
				}

				$events[] = $event;
			}

			return $events;
		} catch (Exception $e) {
			$this->log($e);
		}
	}

	protected function getIcalUrl($calendar)
	{
		return 'https://calendar.google.com/calendar/ical/' . $calendar->params->get('calendarId') . '/public/basic.ics';
	}

	public function saveEvent($eventId, $calendarId, array $data)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return '';
		}

		if (!empty($data['description']) && strlen($data['description']) > 8192) {
			throw new InvalidArgumentException(Text::sprintf('PLG_DPCALENDAR_GOOGLE_ERROR_DESCRIPTION_LENGTH', strlen($data['description'])));
		}

		try {
			$event = new stdClass();
			if (!empty($eventId)) {
				$event = $this->fetchEventFromGoogle(true, $eventId, $calendarId);
			}

			$event->summary     = $data['title'];
			$event->description = !empty($data['description']) ? $data['description'] : '';
			if (isset($data['location_ids'])) {
				$event->location = Location::format(Location::getLocations((array)$data['location_ids']));
			}

			$allDay                 = $data['all_day'] == '1';
			$startDate              = DPCalendarHelper::getDate($data['start_date'], $allDay);
			$event->start           = new stdClass();
			$event->start->timeZone = $startDate->getTimezone()->getName();
			if ($allDay) {
				$event->start->date = $startDate->format('Y-m-d');
			} else {
				$event->start->dateTime = $startDate->format('c');
			}

			$endDate              = DPCalendarHelper::getDate($data['end_date'], $allDay);
			$event->end           = new stdClass();
			$event->end->timeZone = $endDate->getTimezone()->getName();
			if ($allDay) {
				$event->end->date = $endDate->format('Y-m-d');
			} else {
				$event->end->dateTime = $endDate->format('c');
			}

			if (isset($data['rrule']) && $data['rrule']) {
				$event->recurrence = ['RRULE:' . $data['rrule']];
			}

			$url = '/calendars/' . $calendar->params->get('calendarId') . '/events';
			if (!empty($event->id)) {
				$url .= '/' . $event->id;
			}
			$event = $this->callAPI(
				$calendar,
				'/calendars/' . urlencode($calendar->params->get('calendarId')) . '/events' . (!empty($event->id) ? '/' . $event->id : ''),
				[],
				json_encode($event),
				empty($eventId) ? 'post' : 'patch'
			);

			$id = $event->id . '_' . ($allDay ? $startDate->format('Ymd') : $startDate->format('YmdHi'));
			if (!empty($data['rrule'])) {
				$id = $event->id . '_0';
			}

			return $this->createEvent($id, $calendarId)->id;
		} catch (Exception $e) {
			$this->log($e->getMessage());

			return false;
		}
	}

	public function deleteEvent($eventId, $calendarId)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return '';
		}

		try {
			$event  = $this->fetchEventFromGoogle(true, $eventId, $calendarId);
			$result = $this->callAPI(
				$calendar,
				'/calendars/' . urlencode($calendar->params->get('calendarId')) . '/events/' . $event->id,
				[],
				'',
				'delete'
			);

			if (!empty($result->error)) {
				throw new Exception($result->error->message, $result->error->code);
			}

			return true;
		} catch (Exception $e) {
			$this->log($e);

			return false;
		}
	}

	public function prepareForm($eventId, $calendarId, $form, $data)
	{
		$form->removeField('color');
		$form->removeField('url');

		return parent::cleanupFormForEdit($calendarId, $form, $data);
	}

	public function import()
	{
		$session = Factory::getSession(['expire' => 30]);

		// If we are on the API callback from google don't save
		if (!$this->app->input->get('code')) {
			$params = $this->app->input->get('params', ['client-id' => null, 'client-secret' => null], 'array');
			$session->set('client-id', $params['client-id'], $this->_name);
			$session->set('client-secret', $params['client-secret'], $this->_name);
		}

		$clientId     = $session->get('client-id', null, $this->_name);
		$clientSecret = $session->get('client-secret', null, $this->_name);

		if ($this->app->input->get('code')) {
			$session->set('client-id', null, $this->_name);
			$session->set('client-secret', null, $this->_name);
		}

		try {
			$uri = !isset($_SERVER['HTTP_HOST']) ? Uri::getInstance('http://localhost') : Uri::getInstance();
			if (filter_var($uri->getHost(), FILTER_VALIDATE_IP)) {
				$uri->setHost('localhost');
			}
			$uri = $uri->toString([
					'scheme',
					'host',
					'port',
					'path'
				]) . '?option=com_dpcalendar&task=plugin.action&plugin=google&action=import';

			if (!$this->app->input->get('code')) {
				$url = 'https://accounts.google.com/o/oauth2/auth?response_type=code&access_type=offline&approval_prompt=force';
				$url .= '&client_id=' . $clientId;
				$url .= '&redirect_uri=' . urlencode($uri);
				$url .= '&scope=' . urlencode('https://www.googleapis.com/auth/calendar');

				$this->app->redirect($url);
				$this->app->close();
			}
			$token = (new HTTP())->post(
				'https://oauth2.googleapis.com/token',
				http_build_query(
					[
						'grant_type'    => 'authorization_code',
						'client_id'     => $clientId,
						'client_secret' => $clientSecret,
						'redirect_uri'  => $uri,
						'code'          => $this->app->input->get('code', null, null)
					]
				)
			);

			if (!empty($token->error)) {
				throw new Exception($token->error->error_description);
			}

			$calendars = (new HTTP())->get(
				'https://www.googleapis.com/calendar/v3/users/me/calendarList',
				$token->access_token
			);

			if (!empty($calendars->error)) {
				throw new Exception($calendars->error->message, $calendars->error->code);
			}

			$existingCalendars = $this->fetchCalendars();
			foreach ($calendars->items as $cal) {
				$model  = BaseDatabaseModel::getInstance('Extcalendar', 'DPCalendarModel');
				$params = new Registry();
				$params->set('refreshToken', $token->refresh_token);
				$params->set('accessToken', $token->access_token);
				$params->set('client-id', $clientId);
				$params->set('client-secret', $clientSecret);
				$params->set('calendarId', $cal->id);
				$params->set('action-create', true);
				$params->set('action-edit', true);
				$params->set('action-delete', true);

				$data = [];
				foreach ($existingCalendars as $calendar) {
					if ($calendar->params->get('calendarId') === $cal->id) {
						$data['id'] = str_replace($this->identifier . '-', '', $calendar->id);
						break;
					}
				}

				if (empty($data['id'])) {
					$data['title'] = $cal->summary;
					$data['color'] = $cal->backgroundColor;
				}

				$data['params'] = $params->toString();
				$data['plugin'] = 'google';

				if (!$model->save($data)) {
					$this->app->enqueueMessage($model->getError(), 'warning');
				}
			}
		} catch (Exception $e) {
			$this->log($e->getMessage());
		}
		$this->app->redirect(
			$this->app->getSession()->get(
				'extcalendarOrigin',
				'index.php?option=com_dpcalendar&view=extcalendars&dpplugin=google&tmpl=' . $this->app->input->get('tmpl'),
				'DPCalendar'
			)
		);
		$this->app->close();
	}

	private function callAPI($calendar, $path, $params = [], $body = null, $method = 'get')
	{
		$result = (new HTTP())->request(
			'https://www.googleapis.com/calendar/v3' . $path . ($params ? '?' . http_build_query($params) : ''),
			$body,
			$calendar->params->get('accessToken'),
			null,
			[],
			[],
			$method
		);

		if (empty($result->error)) {
			return $result;
		}

		if ($result->error->status != 'PERMISSION_DENIED' && $result->error->status != 'UNAUTHENTICATED') {
			throw new Exception($result->error->message, $result->error->code);
		}

		$token = (new HTTP())->post(
			'https://oauth2.googleapis.com/token',
			http_build_query(
				[
					'grant_type'    => 'refresh_token',
					'refresh_token' => $calendar->params->get('refreshToken'),
					'client_id'     => $calendar->params->get('client-id'),
					'client_secret' => $calendar->params->get('client-secret')
				]
			)
		);

		if (!empty($token->error)) {
			throw new Exception($token->error_description);
		}

		$calendar->params->set('accessToken', $token->access_token);

		$data           = (array)$calendar;
		$data['id']     = str_replace($this->identifier . '-', '', $calendar->id);
		$data['params'] = $calendar->params->toString();

		BaseDatabaseModel::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/models', 'DPCalendarModel');
		$model = BaseDatabaseModel::getInstance('Extcalendar', 'DPCalendarModel', ['ignore_request' => true]);
		$model->save($data);

		$result = (new HTTP())->request(
			'https://www.googleapis.com/calendar/v3' . $path . ($params ? '?' . http_build_query($params) : ''),
			$body,
			$calendar->params->get('accessToken'),
			null,
			[],
			[],
			$method
		);

		if (!empty($result->error)) {
			throw new Exception($result->error->message, $result->error->code);
		}

		return $result;
	}

	private function getDate($date, $allDay, $obj)
	{
		if (!$date) {
			return null;
		}
		$time = empty($date->date) ? $date->dateTime : $date->date;

		$tz = !empty($date->timeZone) ? $date->timeZone : null;
		if (!$tz && isset($obj->timeZone)) {
			$tz = $obj->timeZone;
		}

		$newDate = DPCalendarHelper::getDate($time, $allDay, $allDay ? null : $tz);
		if (!$allDay && $tz) {
			$newDate->setTimezone(new DateTimeZone($tz));
		}

		return $newDate;
	}

	private function createEventFromGoogle($googleEvent, $calendar, $obj, $colors)
	{
		if (!$googleEvent->start) {
			return null;
		}

		$allDay = empty($googleEvent->start->dateTime);

		$startDate = $this->getDate($googleEvent->start, $allDay, $obj);
		$endDate   = $this->getDate($googleEvent->end, $allDay, $obj);
		if ($allDay) {
			$endDate->modify('-1 day');
		}

		$calendarId = $calendar->params->get('calendarId');
		$id         = $googleEvent->id;
		if (strpos($id, '_R') === false &&
			strpos($calendarId, '#sports@group.v.calendar.google.com') === false &&
			strpos($calendarId, '#holiday@group.v.calendar.google.com') === false
		) {
			$pos = strrpos($id, '_');
			if ($pos !== false && $pos !== 0) {
				$id = substr($id, 0, $pos);
			}
		}
		if (!empty($googleEvent->recurringEventId)) {
			$id = $googleEvent->recurringEventId;
		}

		if (!empty($googleEvent->recurrence)) {
			$id .= '_0';
		} else {
			$id .= '_' . ($allDay ? $startDate->format('Ymd') : $startDate->format('YmdHi'));
		}

		$at        = strpos($id, '@');
		$delimiter = strrpos($id, '_');
		if ($at !== false && $delimiter !== false) {
			$id = substr_replace($id, '', $at, $delimiter - $at);
		}

		$event                 = $this->createEvent($id, str_replace('g-', '', $calendar->id));
		$event->uid            = $googleEvent->iCalUID;
		$event->start_date     = $startDate->toSql();
		$event->end_date       = $endDate->toSql();
		$event->all_day        = $allDay ? 1 : 0;
		$event->access_content = $calendar->access_content;
		$event->created        = DPCalendarHelper::getDate($googleEvent->created)->toSql();
		$event->modified       = DPCalendarHelper::getDate($googleEvent->updated)->toSql();
		$event->state          = $googleEvent->status == 'confirmed' ? 1 : -2;

		$event->title = $googleEvent->summary;
		if (!empty($googleEvent->description) && $calendar->params->get('format-description', '1') == '1') {
			$description        = str_replace('\n', '<br/>', $googleEvent->description);
			$event->description = DPCalendarHelper::parseHtml($description);
		} elseif (!empty($googleEvent->description)) {
			$event->description = $googleEvent->description;
		}

		if (!empty($googleEvent->attachments)) {
			$path = PluginHelper::getLayoutPath('dpcalendar', 'google', 'attachments');

			ob_start();
			include $path;
			$event->description .= ob_get_clean();
		}

		// This is the original event
		if (!empty($googleEvent->recurrence)) {
			foreach ($googleEvent->recurrence as $recValue) {
				if (strpos($recValue, 'RRULE') === false) {
					continue;
				}
				$event->rrule = str_replace('RRULE:', '', $recValue);
			}
			$event->original_id = -1;
		} elseif (!empty($googleEvent->recurringEventId)) {
			$originalAllDay       = empty($googleEvent->originalStartTime->dateTime);
			$recDate              = $this->getDate($googleEvent->originalStartTime, $originalAllDay, $obj);
			$event->recurrence_id = $originalAllDay ? $recDate->format('Ymd') : $recDate->format('Ymd\THis\Z');
			$event->original_id   = substr($event->id, 0, strrpos($event->id, '_')) . '_0';
		}
		if (!empty($googleEvent->location)) {
			$event->locations = [Location::get($googleEvent->location)];
		}

		if (!empty($googleEvent->creator->displayName)) {
			$event->created_by_alias = str_replace('MAILTO:', '', $googleEvent->creator->displayName);
		}

		$event->url = $googleEvent->htmlLink;
		if (!empty($googleEvent->colorId) && $googleEvent->colorId > 0 && !$calendar->color_force) {
			$event->color = str_replace('#', '', $colors->{$googleEvent->colorId}->background);
		}

		return $event;
	}
}
