<?php
/**
 * @package   DPCalendar
 * @copyright Copyright (C) 2014 Digital Peak GmbH. <https://www.digital-peak.com>
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 */
defined('_JEXEC') or die();

if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
	return;
}

use DigitalPeak\ThinHTTP as HTTP;
use Joomla\CMS\Date\Date;
use Joomla\Registry\Registry;

class PlgDPCalendarMeetup extends \DPCalendar\Plugin\SyncPlugin
{
	protected $identifier = 'mu';

	/** @var \Joomla\CMS\Application\CMSApplication */
	protected $app;

	public function fetchEvents($calendarId, Date $startDate = null, Date $endDate = null, Registry $options)
	{
		$events = parent::fetchEvents($calendarId, $startDate, $endDate, $options);
		foreach ($events as $event) {
			$event->description = DPCalendarHelper::parseHtml($event->description);
		}

		return $events;
	}

	protected function getContent($calendarId, Date $startDate = null, Date $endDate = null, Registry $options)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return '';
		}

		$text = [];
		try {
			$params = [
				'text_format'    => 'html',
				'limited_events' => 'false',
				'group_urlname'  => trim($calendar->params->get('group')),
				'fields'         => 'event_hosts',
				'status'         => 'cancelled,past,upcoming',
				'has_ended'      => 'true',
				'scroll'         => 'future_or_past'
			];
			if ($startDate) {
				$params['no_earlier_than'] = $startDate->format('Y-m-d\TH:i:s.000');
			}
			if ($endDate) {
				$params['no_later_than'] = $startDate->format('Y-m-d\TH:i:s.000');
			}

			$events = $this->call($calendar, '/events', $params);
			if (!empty($events->errors)) {
				$this->log($events->errors[0]->message);

				return '';
			}

			// Iterate over each event we received
			$text[] = 'BEGIN:VCALENDAR';
			foreach ($events->results as $event) {
				$text = array_merge($text, $this->event($calendar, $startDate, $endDate, $event));
			}

			// Keep processing additional events we may have, if there's too many they won't be in the original response
			while ($events->meta->next) {
				$events = $this->call($calendar, str_replace('https://api.meetup.com', '', $events->meta->next));
				foreach ($events->results as $event) {
					$text = array_merge($text, $this->event($calendar, $startDate, $endDate, $event));
				}
			}

			$text[] = 'END:VCALENDAR';
		} catch (Exception $e) {
			$this->log($e->getMessage());

			return '';
		}

		return $text;
	}

	protected function event($calendar, $startDate, $endDate, $event)
	{
		$text = [];

		$text[] = 'BEGIN:VEVENT';

		// All day if it's at least 24 hours or more (in milliseconds)?
		$allDay   = (isset($event->duration) && intval($event->duration) >= 86400000);
		$timezone = 'UTC';

		if (isset($event->timezone)) {
			$timezone = $event->timezone;
		}

		// Pass in seconds for time which is in milliseconds
		$startDate = DPCalendarHelper::getDate(($event->utc_time / 1000), $allDay, $timezone);
		if ($allDay) {
			$text[] = 'DTSTART;VALUE=DATE:' . $startDate->format('Ymd');
		} else {
			$text[] = 'DTSTART:' . $startDate->format('Ymd\THis\Z');
		}

		$endDate = null;
		if (!isset($event->duration)) {
			$endDate = clone $startDate;
			if ($allDay) {
				$endDate->modify('+1 day');
			} else {
				$endDate->setTime(23, 59, 59);
			}
		} else {
			$endDate = clone $startDate;
			$endDate->modify('+' . ($event->duration / 1000) . ' seconds');
		}

		if ($allDay) {
			$text[] = 'DTEND;VALUE=DATE:' . $endDate->format('Ymd');
		} else {
			$text[] = 'DTEND:' . $endDate->format('Ymd\THis\Z');
		}

		// include event hosts
		if (isset($event->event_hosts)) {
			if (!empty($event->event_hosts)) {
				$organizer = 'ORGANIZER:';
				foreach ($event->event_hosts as $host) {
					$organizer .= $host->member_name . ',';
				}
				$organizer = substr($organizer, 0, -1);
				$text[]    = $organizer;
			}
		}

		$text[] = 'TITLE:' . $event->name;
		$text[] = 'UID:' . $event->id;
		$text[] = 'CATEGORIES:' . $calendar->params->get('group', 'Default');
		$text[] = 'SUMMARY:' . $event->name;
		$text[] = 'DESCRIPTION:' . (isset($event->description) ? $this->replaceNl(nl2br($event->description)) : '');

		$text[] = 'X-URL:' . $event->event_url;
		if (isset($event->venue)) {
			$location = $event->venue->name;
			if (isset($event->venue->address_1)) {
				$location .= ',' . $event->venue->address_1;
			}
			if (isset($event->venue->city)) {
				$location .= ',' . $event->venue->city;
			}
			if (isset($event->venue->state)) {
				$location .= ',' . $event->venue->state;
			}
			if (isset($event->venue->zip)) {
				$location .= ',' . $event->venue->zip;
			}
			$text[] = 'LOCATION: ' . $location;
			if (isset($event->venue->lat) && isset($event->venue->lon)) {
				$text[] = 'GEO:' . $event->venue->lat . ';' . $event->venue->lon;
			}
		}

		$text[] = 'END:VEVENT';

		return $text;
	}

	public function import()
	{
		$session = JFactory::getSession(['expire' => 30]);

		// Not on callback, access params
		if (!$this->app->input->get('code')) {
			// Access all of our params and store them in the session for subsequent use
			$params = $this->app->input->get(
				'params',
				[
					'key'      => null,
					'secret'   => null,
					'memberId' => null,
					'status'   => ['past', 'upcoming']
				],
				'array'
			);

			if (!key_exists('status', $params)) {
				$params['status'] = ['past', 'upcoming'];
			}

			$session->set('key', $params['key'], $this->_name);
			$session->set('secret', $params['secret'], $this->_name);
			$session->set('memberId', $params['memberId'], $this->_name);
			$session->set('status', $params['status'], $this->_name);
		}

		// Access the session on subsequent calls to this routine
		$key      = $session->get('key', null, $this->_name);
		$secret   = $session->get('secret', null, $this->_name);
		$memberId = $session->get('memberId', null, $this->_name);
		$status   = $session->get('status', ['past', 'upcoming'], $this->_name);

		// Redirect for authorization steps
		$redirect = JUri::getInstance()->toString([
				'scheme',
				'host',
				'port',
				'path'
			]) . '?option=com_dpcalendar&task=plugin.action&dpplugin=meetup&action=import';

		// We're not on the callback
		if (!$this->app->input->get('code')) {
			// Call the meetup API for authorization
			$params = [
				'response_type' => 'code',
				'client_id'     => $key,
				'redirect_uri'  => $redirect
			];

			$this->app->redirect('https://secure.meetup.com/oauth2/authorize?' . http_build_query($params));
			$this->app->close();

			return;
		}

		// We either are processing the code that we requested OR the subsequent response auth details
		try {
			$params = [
				'client_id'     => trim($key),
				'client_secret' => trim($secret),
				'redirect_uri'  => $redirect,
				'grant_type'    => 'authorization_code',
				'code'          => $this->app->input->get('code')
			];

			$response = (new HTTP())->post(
				'https://secure.meetup.com/oauth2/access',
				http_build_query($params)
			);

			if (!empty($response->error)) {
				throw new Exception($response->error);
			}

			$accessToken  = $response->access_token;
			$refreshToken = $response->refresh_token;

			// Clear out parameters
			$session->set('key', null, $this->_name);
			$session->set('secret', null, $this->_name);
			$session->set('memberId', null, $this->_name);
			$session->set('status', null, $this->_name);

			$groups = (new HTTP())->get('https://api.meetup.com/self/groups', $accessToken);
			if (!empty($groups->errors)) {
				throw new Exception($groups->errors[0]->message);
			}
			foreach ($groups->data as $group) {
				$model  = JModelLegacy::getInstance('Extcalendar', 'DPCalendarModel');
				$params = new JRegistry();
				$params->set('group', $group->urlname);
				$params->set('key', $key);
				$params->set('secret', $secret);
				$params->set('memberId', $memberId);
				$params->set('access_token', $accessToken);
				$params->set('refresh_token', $refreshToken);

				$table = JTable::getInstance('Extcalendar', 'DPCalendarTable');
				$table->load(['title' => $group->urlname]);

				$data = [];
				if ($table->id > 0) {
					$data['id'] = $table->id;
				}
				$data['title']       = $group->urlname;
				$data['color']       = substr(dechex(crc32($data['title'])), 0, 6);
				$data['description'] = $group->description;
				$data['params']      = $params->toString();
				$data['plugin']      = 'meetup';
				if (!$model->save($data)) {
					$this->app->enqueueMessage($model->getError(), 'warning');
				}
			}

			$this->app->redirect(
				JFactory::getSession()->get(
					'extcalendarOrigin',
					'index.php?option=com_dpcalendar&view=extcalendars&dpplugin=meetup&tmpl=' . $this->app->input->get('tmpl'),
					'DPCalendar'
				)
			);
			$this->app->close();
		} catch (Exception $e) {
			$this->app->enqueueMessage($e->getMessage(), 'error');

			return;
		}
	}


	private function call($calendar, $path, $params = [])
	{
		$result = (new HTTP())->get(
			'https://api.meetup.com' . $path . ($params ? '?' . http_build_query($params) : ''),
			$calendar->params->get('access_token')
		);

		if (empty($result->code)) {
			return $result;
		}

		if ($result->code != 'not_authorized') {
			throw new Exception($result->details);
		}

		$token = (new HTTP())->post(
			'https://secure.meetup.com/oauth2/access',
			http_build_query(
				[
					'grant_type'    => 'refresh_token',
					'refresh_token' => $calendar->params->get('refresh_token'),
					'client_id'     => $calendar->params->get('key'),
					'client_secret' => $calendar->params->get('secret')
				]
			)
		);

		$calendar->params->set('access_token', $token->access_token);

		$data           = (array)$calendar;
		$data['id']     = str_replace($this->identifier . '-', '', $calendar->id);
		$data['params'] = $calendar->params->toString();

		\JModelLegacy::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/models', 'DPCalendarModel');
		$model = \JModelLegacy::getInstance('Extcalendar', 'DPCalendarModel', ['ignore_request' => true]);
		$model->save($data);

		$result = (new HTTP())->get(
			'https://api.meetup.com/' . $path . ($params ? '?' . http_build_query($params) : ''),
			$calendar->params->get('access_token')
		);

		if (!empty($result->code)) {
			throw new Exception($result->details);
		}

		return $result;
	}
}
