<?php
/**
 * @package   DPCalendar
 * @copyright Copyright (C) 2020 Digital Peak GmbH. <https://www.digital-peak.com>
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 */

defined('_JEXEC') or die();

if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
	return;
}

use DigitalPeak\ThinHTTP as HTTP;
use DPCalendar\Helper\Location;
use DPCalendar\Plugin\SyncPlugin;
use Joomla\CMS\Date\Date;
use Joomla\CMS\Factory;
use Joomla\CMS\MVC\Model\BaseDatabaseModel;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\Table\Table;
use Joomla\CMS\Uri\Uri;
use Joomla\Registry\Registry;

class PlgDPCalendarMicrosoft extends SyncPlugin
{
	protected $identifier = 'mi';

	/** @var \Joomla\CMS\Application\CMSApplication */
	protected $app;

	private $graphUrl = 'https://graph.microsoft.com/v1.0';

	protected function getSyncToken($calendar)
	{
		$params = [];
		if ($calendar->sync_token == 1) {
			$params['startDateTime'] = DPCalendarHelper::getDate(1)->format('c');
			$params['endDateTime']   = DPCalendarHelper::getDate('2038-12-31')->format('c');
		} else {
			$params['$skiptoken'] = $calendar->sync_token;
		}
		$data = $this->callAPI(
			$calendar,
			'/calendars/' . $calendar->params->get('calendar_id') . '/calendarView/delta',
			$params,
			['Prefer: odata.maxpagesize=2']
		);

		if (!empty($data->{'@odata.nextLink'})) {
			return substr($data->{'@odata.nextLink'}, strpos($data->{'@odata.nextLink'}, '$skiptoken=') + 11);
		}

		return parent::getSyncToken($calendar);
	}

	public function fetchEvent($eventId, $calendarId, $colors = null)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return '';
		}

		try {
			if ($colors === null) {
				$colors = $this->getColorCategories($calendar);
			}

			return $this->createEventFromGraph($this->callAPI($calendar, '/events/' . $eventId), $calendar, $colors);
		} catch (Exception $e) {
			$this->log($e->getMessage());
		}

		return null;
	}

	public function fetchEvents($calendarId, Date $startDate = null, Date $endDate = null, Registry $options)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return '';
		}

		$params = ['$top' => 100];
		if (!$startDate) {
			$startDate = DPCalendarHelper::getDate();
		}
		$params['startDateTime'] = $startDate->format('c');

		if (!$endDate) {
			$endDate = clone $startDate;
			$endDate->modify('+6 months');
		}
		$params['endDateTime'] = $endDate->format('c');

		try {
			$colors         = $this->getColorCategories($calendar);
			$fetchedMasters = [];
			$data           = [];
			while (true) {
				$events = $this->callAPI($calendar, '/calendars/' . $calendar->params->get('calendar_id') . '/calendarView', $params);

				foreach ($events->value as $event) {
					if ($event->seriesMasterId && !$options->get('expand', true) && array_key_exists($event->seriesMasterId, $fetchedMasters)) {
						continue;
					}
					if ($event->seriesMasterId && !$options->get('expand', true)) {
						$data[]                                 = $this->fetchEvent($event->seriesMasterId, $calendarId, $colors);
						$fetchedMasters[$event->seriesMasterId] = $event->seriesMasterId;
						continue;
					}
					$data[] = $this->createEventFromGraph($event, $calendar, $colors);
				}

				if (empty($events->{'@odata.nextLink'})) {
					break;
				}

				$params = Uri::getInstance($events->{'@odata.nextLink'})->getQuery(true);
			}
		} catch (Exception $e) {
			$this->log($e->getMessage());
		}

		return $data;
	}

	public function import()
	{
		$inputParams = $this->app->input->get('params', ['tenant_id' => null, 'client_id' => null, 'client_secret' => null], 'array');
		$inputParams = new Registry($inputParams);

		try {
			$http = new HTTP();

			$token = $http->post(
				'https://login.microsoftonline.com/' . $inputParams->get('tenant_id') . '/oauth2/v2.0/token',
				http_build_query([
					'client_id'     => $inputParams->get('client_id'),
					'client_secret' => $inputParams->get('client_secret'),
					'scope'         => 'https://graph.microsoft.com/.default',
					'grant_type'    => 'client_credentials'
				])
			);
			if (!empty($token->error)) {
				throw new Exception($token->error_description);
			}

			$existingCalendars = $this->fetchCalendars();
			$page              = $this->graphUrl . '/users?$top=999';
			while (true) {
				$users = $http->get($page, $token->access_token);
				if (!empty($users->error)) {
					throw new Exception($users->error->message);
				}

				foreach ($users->value as $user) {
					$calendars = $http->get(
						$this->graphUrl . '/users/' . $user->id . '/calendars',
						$token->access_token
					);
					if (!empty($calendars->error)) {
						$this->app->enqueueMessage(
							'Error on calendars for user ' . $user->displayName . ':' . ($calendars->error->message),
							'warning'
						);
						continue;
					}

					foreach ($calendars->value as $calendar) {
						$model  = BaseDatabaseModel::getInstance('Extcalendar', 'DPCalendarModel');
						$params = new Registry();
						$params->set('tenant_id', $inputParams->get('tenant_id'));
						$params->set('client_id', $inputParams->get('client_id'));
						$params->set('client_secret', $inputParams->get('client_secret'));
						$params->set('calendar_id', $calendar->id);
						$params->set('user_id', $user->id);
						$params->set('access_token', $token->access_token);

						$data = [];
						foreach ($existingCalendars as $cal) {
							// When the same update it
							if ($cal->params->get('calendar_id') === $calendar->id && $cal->params->get('user_id') === $user->id) {
								$data['id'] = str_replace($this->identifier . '-', '', $cal->id);
								break;
							}
						}

						$data['title']  = $user->displayName . ' [' . $calendar->name . ']';
						$data['color']  = $calendar->hexColor ?: substr(dechex(crc32($data['title'])), 0, 6);
						$data['params'] = $params->toString();
						$data['plugin'] = 'microsoft';

						if (!$model->save($data)) {
							$this->app->enqueueMessage($model->getError(), 'warning');
						}
					}
				}

				if (empty($users->{'@odata.nextLink'})) {
					break;
				}

				$page = $users->{'@odata.nextLink'};
			}

			$this->app->redirect(
				Factory::getSession()->get(
					'extcalendarOrigin',
					'index.php?option=com_dpcalendar&view=extcalendars&dpplugin=microsoft&tmpl=' . $this->app->input->get('tmpl'),
					'DPCalendar'
				)
			);
			$this->app->close();
		} catch (Exception $e) {
			$this->app->enqueueMessage($e->getMessage(), 'warning');
			$this->app->redirect(
				Factory::getSession()->get(
					'extcalendarOrigin',
					'index.php?option=com_dpcalendar&view=extcalendars&dpplugin=microsoft&tmpl=' . $this->app->input->get('tmpl'),
					'DPCalendar'
				)
			);
		}
	}

	private function createEventFromGraph(stdClass $graphEvent, stdClass $calendar, array $colors)
	{
		$startDate = DPCalendarHelper::getDate($graphEvent->start->dateTime, $graphEvent->isAllDay);
		$endDate   = DPCalendarHelper::getDate($graphEvent->end->dateTime, $graphEvent->isAllDay);
		if ($graphEvent->isAllDay) {
			$endDate->modify('-1 day');
		}

		$event                 = $this->createEvent($graphEvent->id, str_replace('mi-', '', $calendar->id));
		$event->uid            = $graphEvent->iCalUId;
		$event->start_date     = $startDate->toSql();
		$event->end_date       = $endDate->toSql();
		$event->all_day        = $graphEvent->isAllDay ? 1 : 0;
		$event->access_content = $calendar->access_content;
		$event->created        = DPCalendarHelper::getDate($graphEvent->createdDateTime)->toSql();
		$event->modified       = DPCalendarHelper::getDate($graphEvent->lastModifiedDateTime)->toSql();
		$event->state          = !$graphEvent->isCancelled ? 1 : 3;
		$event->url            = $graphEvent->webLink;

		$event->title = $graphEvent->subject;
		if ($graphEvent->body->content) {
			preg_match("/<body[^>]*>(.*?)<\/body>/is", $graphEvent->body->content, $matches);
			$event->description = $matches ? $matches[1] : $graphEvent->body->content;
		}

		$event->color = $calendar->color;
		if (!$calendar->color_force && $graphEvent->categories && array_key_exists($graphEvent->categories[0], $colors)) {
			$event->color = $colors[$graphEvent->categories[0]];
		}

		if ($graphEvent->hasAttachments) {
			$this->storeAttachments($event, $graphEvent, $calendar);
		}

		if (!empty($graphEvent->recurrence)) {
			$event->rrule = '';

			switch ($graphEvent->recurrence->pattern->type) {
				case 'daily':
					$event->rrule = 'FREQ=DAILY;';
					break;
				case 'weekly':
					$event->rrule = 'FREQ=WEEKLY;';
					if (!empty($graphEvent->recurrence->pattern->daysOfWeek)) {
						$event->rrule .= 'BYDAY=';
						foreach ($graphEvent->recurrence->pattern->daysOfWeek as $day) {
							$event->rrule .= strtoupper(substr($day, 0, 2)) . ',';
						}
						$event->rrule = trim($event->rrule, ',') . ';';
					}
					break;
				case 'absoluteMonthly':
					$event->rrule = 'FREQ=MONTHLY;';
					if (!empty($graphEvent->recurrence->pattern->dayOfMonth)) {
						$event->rrule .= 'BYMONTHDAY=' . $graphEvent->recurrence->pattern->dayOfMonth . ';';
					}
					break;
				case 'relativeMonthly':
					$event->rrule = 'FREQ=MONTHLY;';
					$event->rrule .= 'FREQ=MONTHLY;';
					if (!empty($graphEvent->recurrence->pattern->daysOfWeek)) {
						$event->rrule .= 'BYDAY=';
						foreach ($graphEvent->recurrence->pattern->daysOfWeek as $day) {
							$indexes = ['First' => 1, 'Second' => 2, 'Third' => 3, 'Fourth' => 4, 'Last' => -1];
							$event->rrule .= $indexes[$graphEvent->recurrence->pattern->index] . strtoupper(substr($day, 0, 2)) . ',';
						}
						$event->rrule = trim($event->rrule, ',') . ';';
					}
					break;
				case 'absoluteYearly':
					$event->rrule = 'FREQ=YEARLY;';
					if (!empty($graphEvent->recurrence->pattern->dayOfMonth)) {
						$event->rrule .= 'BYMONTHDAY=' . $graphEvent->recurrence->pattern->dayOfMonth . ';';
						$indexes = [
							'January'   => 1,
							'February'  => 2,
							'March'     => 3,
							'April'     => 4,
							'May'       => 5,
							'June'      => 6,
							'July'      => 7,
							'August'    => 8,
							'September' => 9,
							'October'   => 10,
							'November'  => 11,
							'December'  => 12
						];
						$event->rrule .= 'BYMONTH=' . $indexes[$graphEvent->recurrence->pattern->month] . ';';
					}
					break;
				case 'relativeYearly':
					$event->rrule = 'FREQ=YEARLY;';
					if (!empty($graphEvent->recurrence->pattern->daysOfWeek)) {
						$indexes = ['First' => 1, 'Second' => 2, 'Third' => 3, 'Fourth' => 4, 'Last' => -1];
						$event->rrule .= 'BYDAY=' . $indexes[$graphEvent->recurrence->pattern->index] .
							strtoupper(substr($graphEvent->recurrence->pattern->daysOfWeek, 0, 2)) . ';';

						$indexes = [
							'January'   => 1,
							'February'  => 2,
							'March'     => 3,
							'April'     => 4,
							'May'       => 5,
							'June'      => 6,
							'July'      => 7,
							'August'    => 8,
							'September' => 9,
							'October'   => 10,
							'November'  => 11,
							'December'  => 12
						];
						$event->rrule .= 'BYMONTH=' . $indexes[$graphEvent->recurrence->pattern->month] . ';';
					}
					break;
			}

			if (!empty($graphEvent->recurrence->pattern->interval)) {
				$event->rrule .= 'Interval=' . $graphEvent->recurrence->pattern->interval . ';';
			}

			if ($graphEvent->recurrence->range->type == 'numbered') {
				$event->rrule .= 'COUNT=' . $graphEvent->recurrence->range->numberOfOccurrences . ';';
			}
			if ($graphEvent->recurrence->range->type == 'endDate') {
				$event->rrule .= 'UNTIL=' . DPCalendarHelper::getDate($graphEvent->recurrence->range->endDate)->format('Ymd') . ';';
			}

			$event->original_id = -1;
		}

		if ($graphEvent->type == 'occurrence') {
			$event->recurrence_id = $event->all_day ? $startDate->format('Ymd') : $startDate->format('Ymd\THis\Z');
			$event->original_id   = $graphEvent->seriesMasterId;
		}

		// Check if a location exists
		if (empty($graphEvent->location)) {
			return $event;
		}

		// When plain text address, then type is default
		if (empty($graphEvent->location->address) && $graphEvent->location->locationType === 'default' && !empty($graphEvent->location->displayName)) {
			$event->locations = [Location::get($graphEvent->location->displayName, ',')];
			return $event;
		}

		// When real address, then attribute address must exist
		if (empty($graphEvent->location->address)) {
			return $event;
		}

		$location = '';
		if (!empty($graphEvent->location->address->street)) {
			$location .= $graphEvent->location->address->street . ',';
		}
		if (!empty($graphEvent->location->address->city)) {
			if (!empty($graphEvent->location->address->postalCode)) {
				$location .= ' ' . $graphEvent->location->address->postalCode ;
			}
			$location .= $graphEvent->location->address->city;
			$location .= ',';
		}
		if (!empty($graphEvent->location->address->countryOrRegion)) {
			$location .= $graphEvent->location->address->countryOrRegion . ',';
		}
		if ($location) {
			$event->locations = [Location::get(rtrim($location, ','))];
		}

		return $event;
	}

	private function callAPI($calendar, $path, $params = [], $headers = [])
	{
		$result = (new HTTP())->get(
			$this->graphUrl . '/users/' . $calendar->params->get('user_id') . $path . '?' . http_build_query($params),
			$calendar->params->get('access_token'),
			null,
			$headers
		);

		if (empty($result->error)) {
			return $result;
		}

		if ($result->error->code != 'InvalidAuthenticationToken') {
			throw new Exception($result->error->message);
		}

		$query = [
			'client_id'     => $calendar->params->get('client_id'),
			'client_secret' => $calendar->params->get('client_secret'),
			'scope'         => 'https://graph.microsoft.com/.default',
			'grant_type'    => 'client_credentials'
		];
		$url   = 'https://login.microsoftonline.com/' . $calendar->params->get('tenant_id') . '/oauth2/v2.0/token';
		$token = (new HTTP())->post($url, http_build_query($query));

		if (!empty($token->error_description)) {
			throw new Exception($token->error_description);
		}

		$calendar->params->set('access_token', $token->access_token);

		$data           = (array)$calendar;
		$data['id']     = str_replace($this->identifier . '-', '', $calendar->id);
		$data['params'] = $calendar->params->toString();

		BaseDatabaseModel::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/models', 'DPCalendarModel');
		$model = BaseDatabaseModel::getInstance('Extcalendar', 'DPCalendarModel', ['ignore_request' => true]);
		$model->save($data);

		$result = (new HTTP())->get(
			$this->graphUrl . '/users/' . $calendar->params->get('user_id') . $path . '?' . http_build_query($params),
			$calendar->params->get('access_token'),
			null,
			$headers
		);

		if (!empty($result->error)) {
			throw new Exception($result->error->message);
		}

		return $result;
	}

	private function storeAttachments($event, $graphEvent, $calendar)
	{
		$attachmentsData = $this->callAPI($calendar, '/events/' . $graphEvent->id . '/attachments');

		$cacheDir = JPATH_ROOT . '/cache/plg_dpcalendar_microsoft/attachments/';
		if (!is_dir($cacheDir)) {
			mkdir($cacheDir, 0777, true);
		}

		$attachments = [];
		foreach ($attachmentsData->value as $attachment) {
			if (!$attachment->name) {
				$attachment->name = md5($attachment->id) . '.png';
			}

			$path = $cacheDir . $attachment->name;
			if (!file_exists($path)) {
				$ext = strtolower(pathinfo($path, PATHINFO_EXTENSION));
				if (!$ext) {
					$ext = 'png';
					$path .= '.' . $ext;
					$attachment->name .= '.' . $ext;
				}

				if (in_array($ext, ['jpg', 'jpeg', 'png']) && $im = imagecreatefromstring(base64_decode($attachment->contentBytes))) {
					switch ($ext) {
						case 'png':
							@imagepng($im, $path);
							break;
						case 'jpg':
						case 'jpeg':
							@imagejpeg($im, $path);
					}
				} else {
					file_put_contents($path, base64_decode($attachment->contentBytes));
				}
			}

			$attachments[$attachment->id] = Uri::root() . 'cache/plg_dpcalendar_microsoft/attachments/' . $attachment->name;
		}

		$path = PluginHelper::getLayoutPath('dpcalendar', 'microsoft', 'attachments');

		ob_start();
		include $path;
		$event->description .= ob_get_clean();
	}

	private function getColorCategories(stdClass $calendar)
	{
		if ($calendar->color_force) {
			return [];
		}

		$colors = require PluginHelper::getLayoutPath('dpcalendar', 'microsoft', 'colors');

		$categories = $this->callAPI($calendar, '/outlook/masterCategories', ['$top' => 100]);

		$data = [];
		foreach ($categories->value as $category) {
			$colorIndex = str_replace('preset', '', $category->color);
			if (!array_key_exists($colorIndex, $colors)) {
				continue;
			}
			$data[$category->displayName] = $colors[$colorIndex];
		}

		return $data;
	}
}
