<?php
/**
 * @package   DPCalendar
 * @copyright Copyright (C) 2014 Digital Peak GmbH. <https://www.digital-peak.com>
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 */
defined('_JEXEC') or die();

if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
	return;
}

class PlgDPCalendarPayPayPal extends \DPCalendar\Plugin\PaymentPlugin
{
	protected function startTransaction($booking, $paymentProvider)
	{
		$url = 'https://api-m.' . ($paymentProvider->sandbox ? 'sandbox.' : '') . 'paypal.com';

		$response = $this->getHTTP()->post(
			$url . '/v1/oauth2/token',
			'grant_type=client_credentials',
			trim($paymentProvider->rest_client_id),
			trim($paymentProvider->rest_secret)
		);

		if ($response->dp->info->http_code >= 400) {
			throw new Exception(!empty($response->message) ? $response->message : 'Could not fetch PayPal access token!');
		}

		if (!$response->access_token) {
			throw new Exception('No PayPal access token available!');
		}

		$params = $this->getPurchaseParameters($booking);

		$data = [
			'intent'        => 'sale',
			'payer'         => ['payment_method' => 'paypal'],
			'transactions'  => [
				[
					'amount'         => ['total' => $params['amount'], 'currency' => $params['currency']],
					'description'    => $params['description'],
					'invoice_number' => $paymentProvider->sandbox ? rand() : $booking->uid
				]
			],
			'redirect_urls' => [
				'return_url' => $params['returnUrl'],
				'cancel_url' => $params['cancelUrl']
			]
		];

		$response = $this->getHTTP()->post(
			$url . '/v1/payments/payment',
			json_encode($data),
			$response->access_token
		);

		if ($response->dp->info->http_code >= 400) {
			throw new Exception(!empty($response->message) ? $response->message : 'Could not create PayPal transaction!');
		}

		if ($response->state != 'created') {
			throw new Exception('Could not create PayPal transaction!');
		}

		// Store the transaction id in the booking, is needed on the callback
		if (!$booking->transaction_id && !empty($response->id)) {
			// Get the booking table
			\JTable::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_dpcalendar/tables');
			$bookingTable = \JTable::getInstance('Booking', 'DPCalendarTable');

			// Load the booking so we can store the transaction id
			$bookingTable->load($booking->id);
			$bookingTable->transaction_id = $response->id;
			$bookingTable->store();
		}

		$this->app->redirect($response->links[1]->href);
	}

	protected function finishTransaction($data, $booking, $paymentProvider)
	{
		$url = 'https://api-m.' . ($paymentProvider->sandbox ? 'sandbox.' : '') . 'paypal.com';

		$response = $this->getHTTP()->post(
			$url . '/v1/oauth2/token',
			'grant_type=client_credentials',
			trim($paymentProvider->rest_client_id),
			trim($paymentProvider->rest_secret)
		);

		if ($response->dp->info->http_code >= 400) {
			throw new Exception(!empty($response->message) ? $response->message : 'Could not fetch PayPal access token!');
		}

		if (!$response->access_token) {
			throw new Exception('No PayPal access token available!');
		}

		// Check if we land here because of a cancel action
		if (empty($data['paymentId']) && empty($data['PayerID'])) {
			$response = $this->getHTTP()->get(
				$url . '/v1/payments/payment/' . $booking->transaction_id,
				$response->access_token
			);
		} else {
			$response = $this->getHTTP()->post(
				$url . '/v1/payments/payment/' . $data['paymentId'] . '/execute',
				json_encode(['payer_id' => $data['PayerID']]),
				$response->access_token
			);
		}

		if ($response->dp->info->http_code >= 400) {
			throw new Exception(!empty($response->message) ? $response->message : 'Error while fetching the transaction data from PayPal!');
		}

		// We probably came here from a cancel action
		if ($response->state == 'created') {
			return ['state' => 6];
		}

		$updates = [];

		// Determine the state
		switch ($response->state) {
			case 'created':
				$updates['state'] = 4;
				break;
			case 'approved':
				$updates['state'] = 1;
				break;
			case 'failed':
				$updates['state'] = 6;
				break;
		}

		// Set payer and transaction information
		$updates['transaction_id'] = $response->id;
		if (!empty($response->payer->payer_info)) {
			$updates['payer_id']    = $response->payer->payer_info->payer_id;
			$updates['payer_email'] = $response->payer->payer_info->email;
		}

		if (!empty($response->transactions)) {
			$tr = $response->transactions[0];
			if (!empty($tr->related_resources)) {
				$s                       = $tr->related_resources[0]->sale;
				$updates['txn_type']     = $s->payment_mode;
				$updates['txn_currency'] = $s->amount->currency;
				$updates['gross_amount'] = $s->amount->total;
				$updates['payment_fee']  = $s->transaction_fee->value;
				$updates['net_amount']   = $updates['gross_amount'] - $updates['payment_fee'];
			}
		}
		$updates['raw_data'] = json_encode($response);

		return $updates;
	}

	protected function getPurchaseParameters($booking)
	{
		$purchaseParameters = parent::getPurchaseParameters($booking);

		// Set the description based on the tickets
		if ($booking->tickets) {
			$ticket                            = reset($booking->tickets);
			$purchaseParameters['description'] = '[' . $booking->id . '] ' . $ticket->event_title;
		}

		// Paypal uses cancel url also for return after successfull payment, see #8422
		$purchaseParameters['cancelUrl'] = $purchaseParameters['returnUrl'];

		return $purchaseParameters;
	}
}
