<?php
/**
 * @package   DPCalendar
 * @copyright Copyright (C) 2022 Digital Peak GmbH. <https://www.digital-peak.com>
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 */

defined('_JEXEC') or die();

use chillerlan\QRCode\QRCode;
use DPCalendar\Booking\Stages\CollectEventsAndTickets;
use DPCalendar\Booking\Stages\SetupForMail;
use DPCalendar\Helper\Booking;
use DPCalendar\Helper\DateHelper;
use DPCalendar\Helper\DPCalendarHelper;
use DPCalendar\Plugin\PaymentPlugin;
use DPCalendar\Translator\Translator;
use DPCalendarPay\QR\SendQRCodeInvoice;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\MVC\Model\BaseDatabaseModel;
use Joomla\CMS\Plugin\PluginHelper;
use League\Pipeline\PipelineBuilder;
use Sprain\SwissQrBill\DataGroup\Element\CombinedAddress;
use Sprain\SwissQrBill\DataGroup\Element\CreditorInformation;
use Sprain\SwissQrBill\DataGroup\Element\PaymentAmountInformation;
use Sprain\SwissQrBill\DataGroup\Element\PaymentReference;
use Sprain\SwissQrBill\DataGroup\Element\StructuredAddress;
use Sprain\SwissQrBill\PaymentPart\Output\HtmlOutput\HtmlOutput;
use Sprain\SwissQrBill\QrBill;
use Sprain\SwissQrBill\Reference\QrPaymentReferenceGenerator;

if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
	return;
}

class PlgDPCalendarPayQR extends PaymentPlugin
{
	protected function getInvoiceData($booking, $paymentProvider): array
	{
		\JLoader::import('dpcalendarpay.qr.vendor.autoload', JPATH_PLUGINS);

		$qrBill = $this->getQRCodeBill($paymentProvider, $booking);
		$output = new HtmlOutput($qrBill, substr(DPCalendarHelper::getFrLanguage(), 0, 2));

		$variables                    = [];
		$variables['qrcode']          = $output->setPrintable(false)->setQrCodeImageFormat(QRCode::OUTPUT_IMAGE_PNG)->getPaymentPart();
		$variables['ibanNumber']      = $paymentProvider->swissqr_iban;
		$variables['referenceNumber'] = $qrBill->getPaymentReference()->getFormattedReference();
		$variables['paymentDueDate']  = DPCalendarHelper::getDate($booking->book_date, true)->modify('+30 days')
			->format(DPCalendarHelper::getComponentParameter('event_date_format', 'd.m.Y'));

		$tickets    = $booking->tickets;
		$translator = new Translator();
		$dateHelper = new DateHelper();
		$params     = ComponentHelper::getParams('com_dpcalendar');

		ob_start();
		require PluginHelper::getLayoutPath('dpcalendarpay', 'qr', 'invoice');
		$details = ob_get_clean();

		$fileName = Booking::createPDF($details, $booking->uid, true);
		if (file_exists($fileName)) {
			return ['name' => $booking->uid . '.pdf', 'content' => file_get_contents($fileName)];
		}

		return parent::getInvoiceData($booking, $paymentProvider);
	}

	protected function finishTransaction($data, $booking, $provider)
	{
		// Only send out a mail when state is confirmed, can happen for waiting lists
		if ($booking->state == 3 && $this->app->get('mailonline', false)) {
			\JLoader::import('dpcalendarpay.qr.vendor.autoload', JPATH_PLUGINS);

			$builder = new PipelineBuilder();
			$builder->add(new CollectEventsAndTickets(
				Factory::getApplication(),
				BaseDatabaseModel::getInstance('Booking', 'DPCalendarModel', ['ignore_request' => true])
			));
			$builder->add(new SetupForMail(Factory::getApplication(), ComponentHelper::getParams('com_dpcalendar')));
			$builder->add(new SendQRCodeInvoice(Factory::getMailer()));

			try {
				$qrBill = $this->getQRCodeBill($provider, $booking);
				$output = new HtmlOutput($qrBill, substr(DPCalendarHelper::getFrLanguage(), 0, 2));

				$builder->build()->process(
					(object)[
						'oldItem'         => $booking,
						'item'            => $booking,
						'language'        => $this->app->getLanguage(),
						'qrcode'          => $output->setPrintable(false)->setQrCodeImageFormat(QRCode::OUTPUT_IMAGE_PNG)->getPaymentPart(),
						'ibanNumber'      => $provider->swissqr_iban,
						'referenceNumber' => $qrBill->getPaymentReference()->getFormattedReference()
					]
				);
			} catch (Exception $e) {
				$violations = '';
				foreach ($qrBill->getViolations() as $violation) {
					$violations .= '[' . $violation->getPropertyPath() . '] ' . $violation->getMessage() .  '<br>';
				}
				throw new Exception($violations);
			}
		}

		// Set state to needs payment and some other fields
		$updates = [
			'state'          => !empty($provider->state) ? $provider->state : 4,
			'payer_email'    => $booking->email,
			'transaction_id' => md5($booking->id . 'dpcalendar'),
			'invoice'        => 1
		];

		return $updates;
	}

	private function getQRCodeBill($provider, $booking): QrBill
	{
		$qrBill = QrBill::create();
		$qrBill->setCreditor(CombinedAddress::create(
			$provider->swissqr_name,
			$provider->swissqr_street,
			$provider->swissqr_city,
			$provider->swissqr_country
		));
		$qrBill->setCreditorInformation(CreditorInformation::create($provider->swissqr_iban));
		$qrBill->setUltimateDebtor(StructuredAddress::createWithStreet(
			$booking->name,
			$booking->street,
			$booking->number,
			$booking->zip,
			$booking->city,
			$booking->country_code
		));
		$qrBill->setPaymentAmountInformation(
			PaymentAmountInformation::create(DPCalendarHelper::getComponentParameter('currency', 'USD'), $booking->price)
		);
		$qrBill->setPaymentReference(PaymentReference::create(
			PaymentReference::TYPE_QR,
			QrPaymentReferenceGenerator::generate($provider->swissqr_besr, $booking->id)
		));

		return $qrBill;
	}
}
