<?php
/**
 * @package   DPCalendar
 * @copyright Copyright (C) 2021 Digital Peak GmbH. <https://www.digital-peak.com>
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 */
defined('_JEXEC') or die();

if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
	return;
}

use DPCalendar\Helper\ExportTrait;
use DPCalendar\Plugin\SyncPlugin;
use Joomla\CMS\Date\Date;
use Joomla\CMS\Toolbar\Toolbar;
use Joomla\CMS\Toolbar\ToolbarHelper;
use Joomla\Input\Input;
use Joomla\Registry\Registry;
use PhpOffice\PhpSpreadsheet\Spreadsheet;
use PhpOffice\PhpSpreadsheet\Writer\Xlsx;

class PlgDPCalendarSpreadsheet extends SyncPlugin
{
	use ExportTrait;

	protected $identifier = 'sp';

	/** @var \Joomla\CMS\Application\CMSApplication */
	protected $app;

	public function onDPCalendarAfterToolbar(string $viewName, Toolbar $toolbar)
	{
		if (!in_array($viewName, ['events', 'bookings', 'tickets'])) {
			return;
		}
		ToolbarHelper::custom($viewName . '.xlsxexport', 'download', '', 'PLG_DPCALENDAR_SPREADSHEET_EXPORT_XLSX', false);
	}

	public function onDPCalendarBeforeExecute(Input $input)
	{
		$task = $input->get('task', '');
		if (strpos($task, 'xlsxexport') === false) {
			return;
		}

		list($action, $task) = explode('.', $task);

		$data = [];
		switch ($action) {
			case 'events':
				$data = $this->getEventData();
				break;
			case 'bookings':
				$data = $this->getBookingsData();
				break;
			case 'tickets':
				$data = $this->getTicketsData();
				break;
		}

		\JLoader::import('dpcalendar.spreadsheet.vendor.autoload', JPATH_PLUGINS);

		$spreadsheet = new Spreadsheet();
		$sheet       = $spreadsheet->getActiveSheet();

		foreach ($data as $rowIndex => $row) {
			foreach ($row as $cellIndex => $value) {
				$sheet->setCellValueByColumnAndRow($cellIndex + 1, $rowIndex + 1, $value);
			}
		}

		$writer = new Xlsx($spreadsheet);
		header('Content-Type: application/xlsx');
		header('Content-Disposition: attachement; filename="' . $action . '-' . date('YmdHi') . '.xlsx";');
		$writer->save('php://output');

		$this->app->close();
	}

	protected function getContent($calendarId, Date $startDate = null, Date $endDate = null, Registry $options)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return '';
		}

		/** @var \Joomla\Registry\Registry $params */
		$params = $calendar->params;

		try {
			JLoader::import('dpcalendar.spreadsheet.vendor.autoload', JPATH_PLUGINS);

			$path = $params->get('uri', '');
			if (strpos($path, '/') !== 0) {
				$path = JPATH_ROOT . '/' . $path;
			}

			$k = [
				$params->get('all_day', 'all_day') => $params->get('all_day', 'all_day'),
				$params->get('start_date', 'start_date') => $params->get('start_date', 'start_date'),
				$params->get('end_date', 'end_date') => $params->get('end_date', 'end_date'),
				$params->get('show_end_time', 'show_end_time') => $params->get('show_end_time', 'show_end_time'),
				$params->get('title', 'title') => $params->get('title', 'title'),
				$params->get('description', 'description') => $params->get('description', 'description'),
				$params->get('rrule', 'rrule') => $params->get('rrule', 'rrule'),
				$params->get('location', 'location') => $params->get('location', 'location'),
				$params->get('alias', 'alias') => $params->get('alias', 'alias'),
				$params->get('color', 'color') => $params->get('color', 'color')
			];

			$spreadsheet = \PhpOffice\PhpSpreadsheet\IOFactory::load($path);
			$workSheet   = $spreadsheet->getActiveSheet();

			$text   = [];
			$text[] = 'BEGIN:VCALENDAR';
			foreach ($workSheet->getRowIterator() as $index => $row) {
				$data = $workSheet->rangeToArray('A' . $index . ':' . $workSheet->getHighestColumn() . $index)[0];
				if ($index == 1 && $params->get('has_header', true)) {
					$k = array_flip($data);
					continue;
				}

				try {
					$text[] = 'BEGIN:VEVENT';

					$allDay = $this->getValue($calendar, $data, $k, 'all_day') == 1;

					$startDate = DPCalendarHelper::getDate($this->getValue($calendar, $data, $k, 'start_date'), $allDay);
					if ($allDay) {
						$text[] = 'DTSTART;VALUE=DATE:' . $startDate->format('Ymd');
					} else {
						$text[] = 'DTSTART:' . $startDate->format('Ymd\THis\Z');
					}
					$endDate = DPCalendarHelper::getDate($this->getValue($calendar, $data, $k, 'end_date'), $allDay);
					if ($allDay) {
						$text[] = 'DTEND;VALUE=DATE:' . $endDate->format('Ymd');
					} else {
						$text[] = 'DTEND:' . $endDate->format('Ymd\THis\Z');
					}

					$text[] = 'UID:' . md5($this->getValue($calendar, $data, $k, 'title') . 'spreadsheet');
					$text[] = 'CATEGORIES:' . $this->params->get('title-' . $calendarId, 'Default');
					$text[] = 'SUMMARY:' . $this->getValue($calendar, $data, $k, 'title');
					$text[] = 'DESCRIPTION:' . str_replace("\n\r", '', $this->getValue($calendar, $data, $k, 'description'));

					$rrule = $this->getValue($calendar, $data, $k, 'rrule', false);
					if (empty($rrule)) {
						// Try out only rule for a more convenient way
						$old = $params->get('rrule');
						$params->set('rrule', 'rule');
						$rrule = $this->getValue($calendar, $data, $k, 'rrule', false);
						$params->set('rrule', $old);
					}
					if (!empty($rrule)) {
						$text[] = 'RRULE:' . $rrule;
					}

					$text[] = 'LOCATION:' . $this->getValue($calendar, $data, $k, 'location', false);

					$text[] = 'X-URL:' . $this->getValue($calendar, $data, $k, 'url', false);
					$text[] = 'X-ALIAS:' . $this->getValue($calendar, $data, $k, 'alias', false);
					$text[] = 'X-COLOR:' . $this->getValue($calendar, $data, $k, 'color', false);
					$text[] = 'X-SHOW-END-TIME:' . $this->getValue($calendar, $data, $k, 'show_end_time', false);

					$text[] = 'END:VEVENT';
				} catch (Exception $e) {
					$this->log($e->getMessage());
				}
			}
			$text[] = 'END:VCALENDAR';

			return $text;
		} catch (\Exception $e) {
			$this->log($e->getMessage());

			return '';
		}
	}

	private function getValue($calendar, $data, $k, $key, $throw = true)
	{
		$column = $calendar->params->get($key, $key);
		if (!key_exists($column, $k)) {
			if ($throw) {
				throw new InvalidArgumentException('Key ' . $key . ' not found to get the value from in the spreadsheet file.');
			}

			return null;
		}
		if (!key_exists($k[$column], $data)) {
			if ($throw) {
				throw new InvalidArgumentException('Key ' . $key . ' not found to get the value from in the spreadsheet file.');
			}

			return null;
		}

		return $data[$k[$column]];
	}
}
