<?php
/**
 * @package   DPCalendar
 * @copyright Copyright (C) 2016 Digital Peak GmbH. <https://www.digital-peak.com>
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 */

defined('_JEXEC') or die();

if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
	return;
}

use DPCalendar\Helper\DPCalendarHelper;
use DPCalendar\Plugin\SyncPlugin;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Date\Date;
use Joomla\Registry\Registry;
use Sabre\VObject\Recur\RRuleIterator;

class PlgDPCalendarSQL extends SyncPlugin
{
	/** @var CMSApplication */
	protected $app;

	protected $identifier = 'sq';

	protected function getContent($calendarId, Date $startDate = null, Date $endDate = null, Registry $options)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return '';
		}

		$params = $calendar->params;

		if (!$startDate) {
			$startDate = DPCalendarHelper::getDate(null, false, $params->get('timezone'));
		}

		if ($params->get('timezone')) {
			// Force the calendar timezone as it comes with the user one
			$startDate->setTimezone(new DateTimeZone($params->get('timezone')));
		}

		if ($endDate && $params->get('timezone')) {
			// Force the calendar timezone as it comes with the user one
			$endDate->setTimezone(new DateTimeZone($params->get('timezone')));
		}

		$text = [];
		try {
			$option = ['prefix' => $this->app->get('dbprefix')];

			$option['driver']   = $params->get('driver', $this->app->get('dbtype'));
			$option['host']     = $params->get('host_name', $this->app->get('host'));
			$option['user']     = $params->get('user', $this->app->get('user'));
			$option['password'] = $params->get('pwd', $this->app->get('password'));
			$option['database'] = $params->get('db_name', $this->app->get('db'));

			$db = JDatabaseDriver::getInstance($option);

			$query = $db->getQuery(true);

			$columns     = [];
			$rruleColumn = null;

			$columns[] = $params->get('id_column', 'id');
			$columns[] = $params->get('title_column', 'title');
			$columns[] = $params->get('start_date_column', 'start_date');
			$columns[] = $params->get('end_date_column', 'end_date');

			if (!empty($params->get('description_column'))) {
				$columns[] = $params->get('description_column');
			}
			if (!empty($params->get('all_day_column'))) {
				$columns[] = $params->get('all_day_column');
			}
			if (!empty($params->get('rrule_column'))) {
				$columns[]   = $params->get('rrule_column');
				$rruleColumn = $params->get('rrule_column');
			}
			if (!empty($params->get('location_column'))) {
				$columns[] = $params->get('location_column');
			}
			if (!empty($params->get('url_column'))) {
				$columns[] = $params->get('url_column');
			}
			if (!empty($params->get('alias_column'))) {
				$columns[] = $params->get('alias_column');
			}
			if (!empty($params->get('color_column'))) {
				$columns[] = $params->get('color_column');
			}
			if (!empty($params->get('image_column'))) {
				$columns[] = $params->get('image_column');
			}

			$query->select($columns);
			$query->from($params->get('table_name'));

			if ($joins = $params->get('joins')) {
				foreach ($joins as $join) {
					$query->join($join->type, $join->table . ' on ' . $join->condition);
				}
			}

			$hasTimeZone     = !empty($params->get('timezone'));
			$startColumn     = $params->get('start_date_column', 'start_date');
			$endColumn       = $params->get('end_date_column', 'end_date');
			$startDateString = $db->quote($startDate->toSql($hasTimeZone));
			$dateCondition   = $startColumn . '  >= ' . $startDateString;
			if ($endDate !== null) {
				$endDateString = $db->quote($endDate->toSql($hasTimeZone));
				$dateCondition = '(' . $endColumn . ' between ' . $startDateString . ' and ' . $endDateString . ' or ' . $startColumn . ' between ' . $startDateString .
					' and ' . $endDateString . ' or (' . $startColumn . ' < ' . $startDateString . ' and ' . $endColumn . ' > ' . $endDateString . '))';
			}

			$now = DPCalendarHelper::getDate(null, false, $params->get('timezone'));
			$dateCondition .= ' or ' . $db->quote($now->toSql($hasTimeZone)) . ' between ' . $startColumn . ' and ' . $endColumn;

			if (!empty($params->get('all_day_column'))) {
				$dateCondition .= ' or (' . $startColumn . '=' . $db->quote($now->format('Y-m-d', $hasTimeZone)) . ' and ' . $params->get('all_day_column') . '=1)';
				$dateCondition .= ' or (' . $endColumn . '=' . $db->quote($now->format('Y-m-d', $hasTimeZone)) . ' and ' . $params->get('all_day_column') . '=1)';
			}

			if ($rruleColumn) {
				$dateCondition .= ' or ' . $db->quoteName($rruleColumn) . ' is not null or ' .$db->quoteName($rruleColumn) . " != ''";
			}

			$query->where('(' . $dateCondition . ')');

			if ($condition = $params->get('where_condition')) {
				$query->where($condition);
			}

			if ($order = $params->get('order')) {
				$query->order($order);
			}

			if ($group = $params->get('group')) {
				$query->group($group);
			}

			$db->setQuery($query, 0, $params->get('limit', 0));

			$rows = $db->loadObjectList();
			if (!is_array($rows)) {
				return [];
			}

			if (!$endDate) {
				$endDate = DPCalendarHelper::getDate('2038-12-31');
			}

			$text[] = 'BEGIN:VCALENDAR';
			foreach ($rows as $row) {
				$eventStartDate = $this->getValue($row, $params->get('start_date_column', 'start_date'), $params);
				$eventEndDate   = $this->getValue($row, $params->get('end_date_column', 'end_date'), $params);

				$allDay = strlen($eventStartDate) === 10 || (strpos($eventStartDate, ' 00:00:00') && strpos($eventEndDate, ' 00:00:00'));
				if (!empty($params->get('all_day_column'))) {
					$allDay = $this->getValue($row, $params->get('all_day_column'), $params) == 1;
				}

				// Check if the start date is not set
				if (!$eventStartDate || $eventStartDate === $db->getNullDate() || $eventStartDate === '0000-00-00') {
					$eventStartDate = null;
				}

				// Set the proper start date
				$eventStartDate = DPCalendarHelper::getDate($eventStartDate, $allDay, !$allDay ? $params->get('timezone') : null);

				// Check if the end date is not set
				if (!$eventEndDate || $eventEndDate === $db->getNullDate() || $eventEndDate === '0000-00-00') {
					$eventEndDate = clone $eventStartDate;

					// When is not an all day, set it to the end of the day
					if (!$allDay) {
						$eventEndDate->setTime(23, 59);
					}
				}

				// Set the proper end date
				$eventEndDate = DPCalendarHelper::getDate($eventEndDate, $allDay, !$allDay ? $params->get('timezone') : null);

				// Set all day when start and end time is not set
				if ($eventStartDate->format('H:i') === '00:00' && $eventEndDate->format('H:i') === '00:00') {
					$allDay = true;
				}

				// Check rrule if an instance is within the time range
				if ($rruleColumn
					&& ($rrule = $this->getValue($row, $params->get('rrule_column'), $params))
					&& $startDate->toUnix() < $endDate->toUnix()) {
					$it = new RRuleIterator($rrule, $eventStartDate);

					// When infinite, then we will find a value anyway
					if (!$it->isInfinite()) {
						$eventLengthInSeconds = $eventEndDate->toUnix() - $eventStartDate->toUnix();
						// The flag if the instance is out of range
						$outOfRange = true;
						foreach ($it as $instanceStartDate) {
							// Add the diff from the event start date
							$instanceEndDate = clone $instanceStartDate;
							$instanceEndDate->modify('+' . $eventLengthInSeconds . ' seconds');

							// If start is after end or end before start, then the event doesn't fit
							if ($instanceStartDate->toUnix() > $endDate->toUnix() || $instanceEndDate->toUnix() < $startDate->toUnix()) {
								continue;
							}
							$outOfRange = false;
							break;
						}

						// When all is out of range, then continue
						if ($outOfRange) {
							continue;
						}
					}
				}

				$text[] = 'BEGIN:VEVENT';

				if ($allDay) {
					$text[] = 'DTSTART;VALUE=DATE:' . $eventStartDate->format('Ymd');
				} else {
					$text[] = 'DTSTART:' . $eventStartDate->format('Ymd\THis\Z');
				}

				if ($allDay && $eventStartDate->format('U') == $eventEndDate->format('U')) {
					$eventEndDate->modify('+1 day');
				}

				if ($allDay) {
					$text[] = 'DTEND;VALUE=DATE:' . $eventEndDate->format('Ymd');
				} else {
					$text[] = 'DTEND:' . $eventEndDate->format('Ymd\THis\Z');
				}

				$text[] = 'UID:' . $this->getValue($row, $params->get('id_column', 'id'), $params);
				$text[] = 'CATEGORIES:Default';
				$text[] = 'SUMMARY:' . $this->getValue($row, $params->get('title_column', 'title'), $params);

				if (!empty($params->get('description_column'))) {
					$text[] = 'DESCRIPTION:' . $this->replaceNl(nl2br($this->getValue($row, $params->get('description_column'), $params)));
				}
				if (!empty($params->get('rrule_column'))
						&& $this->getValue($row, $params->get('rrule_column'), $params)) {
					$text[] = 'RRULE:' . $this->getValue($row, $params->get('rrule_column'), $params);
				}
				if (!empty($params->get('location_column'))) {
					$text[] = 'LOCATION:' . $this->getValue($row, $params->get('location_column'), $params);
				}
				if (!empty($params->get('url_column'))) {
					$text[] = 'X-URL:' . $this->getValue($row, $params->get('url_column'), $params);
				}
				if (!empty($params->get('alias_column'))) {
					$text[] = 'X-ALIAS:' . $this->getValue($row, $params->get('alias_column'), $params);
				}
				if (!empty($params->get('color_column'))) {
					$text[] = 'X-COLOR:' . $this->getValue($row, $params->get('color_column'), $params);
				}
				if (!empty($params->get('image_column'))) {
					$text[] = 'X-IMAGE:' . $this->getValue($row, $params->get('image_column'), $params);
				}

				$text[] = 'END:VEVENT';
			}

			$text[] = 'END:VCALENDAR';
		} catch (Exception $e) {
			$this->log($e->getMessage());

			if (!empty($query) && JDEBUG) {
				$this->log($query);
			}
		}

		return $text;
	}

	/**
	 * Returns the value of a result set row from a DB query. This is needed as the columns in the result set
	 * have no alias. So it is required to strip the alias from the column definition first before getting the value.
	 *
	 * @param \stdClass $data
	 * @param string $column
	 * @param Registry $params
	 *
	 * @return string
	 */
	private function getValue(\stdClass $data, string $column, Registry $params): ?string
	{
		// Collect the tables
		$tables = [$params->get('table_name')];
		if ($joins = $params->get('joins')) {
			foreach ($joins as $join) {
				$tables[] = $join->table;
			}
		}

		// When the column is an alias, then remove it as the data will have it without alias
		$pos = strrpos($column, ' ');
		if ($pos !== false && $pos >= 0) {
			$column = trim(substr($column, $pos));
		}

		// Loop over tables
		foreach ($tables as $table) {
			// Compile the alias
			$tableDefinition = explode(' ', trim($table));
			$tableAlias      = '';
			if (count($tableDefinition) === 2) {
				$tableAlias = $tableDefinition[1] . '.';
			}
			if (count($tableDefinition) === 3) {
				$tableAlias = $tableDefinition[2] . '.';
			}

			if (!$tableAlias) {
				continue;
			}

			// If the alias is part of the column then remove it as the data has the keys without column
			if (strpos($column, $tableAlias) === 0) {
				$column = str_replace($tableAlias, '', $column);
			}
		}

		return $data->{$column};
	}
}
