<?php
/**
 * @package   DPCalendar
 * @copyright Copyright (C) 2020 Digital Peak GmbH. <https://www.digital-peak.com>
 * @license   http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 */

defined('_JEXEC') or die();

if (!JLoader::import('components.com_dpcalendar.helpers.dpcalendar', JPATH_ADMINISTRATOR)) {
	return;
}

use DigitalPeak\ThinHTTP as HTTP;
use DPCalendar\Helper\DPCalendarHelper;
use DPCalendar\HTML\Document\HtmlDocument;
use DPCalendar\Plugin\SyncPlugin;
use DPCalendar\Translator\Translator;
use Firebase\JWT\JWT;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Date\Date;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Model\BaseDatabaseModel;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\Table\Table;
use Joomla\CMS\Uri\Uri;
use Joomla\Registry\Registry;

class PlgDPCalendarZoom extends SyncPlugin
{
	protected $identifier = 'z';

	/** @var CMSApplication */
	protected $app;

	/** @var \Joomla\Database\DatabaseDriver */
	protected $db;

	public function onContentAfterDisplay($context, $event)
	{
		if ($context != 'com_dpcalendar.event') {
			return;
		}

		$params = $event->params;
		if (!$params->get('zoom_meeting_nr') || !$params->get('zoom_calendar')) {
			return;
		}

		$user = Factory::getUser();
		if ($params->get('zoom_only_tickets') && $user->guest) {
			return;
		}

		if ($params->get('zoom_only_tickets') && $event->created_by != $user->id && !$user->authorise('dpcalendar.admin.book', 'com_dpcalendar')) {
			$myTickets = array_filter($event->tickets, function ($ticket) use ($user) {
				return $ticket->user_id == $user->id && $ticket->state == 1;
			});

			if (!$myTickets) {
				return;
			}
		}

		$calendar = $this->getDbCal($params->get('zoom_calendar'));
		if (!$calendar) {
			return;
		}

		$leaveUrl = Uri::current();
		if ($this->app->input->get('option') != 'com_dpcalendar' || $this->app->input->get('view') != 'event') {
			$leaveUrl = DPCalendarHelperRoute::getEventRoute($event->id, $event->catid, true, true);
		}

		$meetingId = preg_replace("/[^0-9]/", '', $params->get('zoom_meeting_nr'));
		$signature = $this->generateSignature(
			$calendar->params->get('api_key'),
			$calendar->params->get('api_secret'),
			$meetingId,
			$event->created_by == $user->id ? 1 : 0
		);

		// When the event has not started yet load the join template
		$path = null;
		if (DPCalendarHelper::getDate()->toSql() < $event->end_date) {
			$path = PluginHelper::getLayoutPath('dpcalendar', 'zoom', 'join');
		} elseif ($params->get('zoom_show_recordings', $calendar->params->get('show_recordings', 1))) {
			$cache = Factory::getCache('plg_' . $this->_type . '_' . $this->_name);
			$cache->setCaching($this->params->get('cache', 1) == '1' && $this->cachingEnabled);
			$cache->setLifeTime($this->params->get('cache_time', 900) / 60);
			$cache->options['locking'] = false;

			$startDate  = DPCalendarHelper::getDate($event->start_date);
			$recordings = null;
			try {
				$recordings = $cache->get([$this, 'getRecordings'], [$calendar->params, $params->get('zoom_meeting_nr'), $startDate]);
				$cache->gc();
			} catch (\Exception $e) {
				$this->log($e->getMessage());
			}

			if ($recordings) {
				$path = PluginHelper::getLayoutPath('dpcalendar', 'zoom', 'recordings');
			}
		}

		if (!$path) {
			return;
		}

		$document   = new HtmlDocument();
		$translator = new Translator();
		$app        = $this->app;

		ob_start();
		include $path;

		return ob_get_clean();
	}

	public function onContentAfterSave($context, $event, $isNew)
	{
		if ($context != 'com_dpcalendar.event') {
			return;
		}

		$params = new Registry($event->params);
		if (!$params->get('zoom_calendar') || ($params->get('zoom_meeting_nr') && $params->get('zoom_meeting_password_hashed'))) {
			return;
		}

		$calendar = $this->getDbCal($params->get('zoom_calendar'));
		if (empty($calendar) || !$calendar->params->get('create_meeting')) {
			return;
		}

		// Set the hashed password when not defined
		if ($params->get('zoom_meeting_nr') && !$params->get('zoom_meeting_password_hashed')) {
			$meeting = $this->talkToAPI($calendar->params, '/meetings/' . preg_replace('/[^0-9]/', '', $params->get('zoom_meeting_nr')));
			if (empty($meeting->password)) {
				return;
			}
			$params->set('zoom_meeting_password', $meeting->password);
			$params->set('zoom_meeting_password_hashed', $meeting->encrypted_password);
			$event->params = $params->toString();

			$this->db->setQuery('update #__dpcalendar_events set params = ' . $this->db->quote($params->toString()) . ' where id = ' . (int)$event->id);
			$this->db->execute();

			return;
		}

		$data             = new stdClass();
		$data->topic      = $event->title;
		$data->type       = 2;
		$data->start_time = DPCalendarHelper::getDate($event->start_date)->format('Y-m-d\TH:i:s\Z');
		$data->duration   = (DPCalendarHelper::getDate($event->end_date)->format('U') - DPCalendarHelper::getDate($event->start_date)->format('U')) / 60;
		$data->password   = $params->get('zoom_meeting_password');
		$data->agenda     = substr(strip_tags($event->description), 0, 2000);

		try {
			$meeting = $this->talkToAPI($calendar->params, 'users/' . $calendar->params->get('user_id') . '/meetings', json_encode($data));
		} catch (Exception $e) {
			$this->log($e->getMessage());

			return;
		}

		$params->set('zoom_meeting_nr', $meeting->id);
		if (!empty($meeting->password)) {
			$params->set('zoom_meeting_password', $meeting->password);
		}
		if (!empty($meeting->encrypted_password)) {
			$params->set('zoom_meeting_password_hashed', $meeting->encrypted_password);
		}

		$event->params = $params->toString();

		$this->db->setQuery('update #__dpcalendar_events set params = ' . $this->db->quote($params->toString()) . ' where id = ' . (int)$event->id);
		$this->db->execute();

		$this->app->enqueueMessage(
			Text::sprintf('PLG_DPCALENDAR_ZOOM_MESSAGE_CREATED_MEETING', 'https://zoom.us/meeting/' . $meeting->id, $meeting->id)
		);
	}

	private function generateSignature($apiKey, $apiSecret, $meetingNumber, $role)
	{
		// See https://marketplace.zoom.us/docs/sdk/native-sdks/web/essential/signature
		$time = time() * 1000 - 30000;//time in milliseconds (or close enough)
		$data = base64_encode($apiKey . $meetingNumber . $time . $role);
		$hash = hash_hmac('sha256', $data, $apiSecret, true);
		$_sig = $apiKey . "." . $meetingNumber . "." . $time . "." . $role . "." . base64_encode($hash);

		//return signature, url safe base64 encoded
		return rtrim(strtr(base64_encode($_sig), '+/', '-_'), '=');
	}

	public function fetchEvent($eventId, $calendarId)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return '';
		}

		$params = $calendar->params;

		$occurrenceId = null;
		if (strpos($eventId, '_')) {
			list($eventId, $occurrenceId) = explode('_', $eventId);
		}

		$meeting = $this->talkToAPI($params, '/meetings/' . $eventId);
		if (!$meeting) {
			return null;
		}

		$meeting->original_id = $meeting->type == 8 ? -1 : 0;

		if ($meeting->type == 8) {
			foreach ($meeting->occurrences as $occurrence) {
				if ($occurrence->occurrence_id != $occurrenceId) {
					continue;
				}

				// Create the instance based on the meeting
				$instance              = $meeting;
				$instance->id          = $meeting->id . '_' . $occurrence->occurrence_id;
				$instance->original_id = $meeting->id;
				$instance->start_time  = $occurrence->start_time;
				$instance->duration    = $occurrence->duration;
				unset($instance->occurrences);
				unset($instance->recurrence);
				$instance->type = 2;

				$meeting = $instance;
			}
		}

		return $this->createEventFromMeeting($meeting, $calendarId);
	}

	public function fetchEvents($calendarId, Date $startDate = null, Date $endDate = null, Registry $options)
	{
		$calendar = $this->getDbCal($calendarId);
		if (empty($calendar)) {
			return '';
		}

		$params = $calendar->params;

		$startDateString = $startDate->format('c');
		$endDateString   = $endDate ? $endDate->format('c') : '';

		$events = [];
		foreach ($this->talkToAPI($params, '/users/' . $params->get('user_id') . '/meetings', '', ['page_size' => 300])->meetings as $meeting) {
			$meetings             = [$meeting];
			$meeting->original_id = 0;
			if ($meeting->type == 8) {
				$meeting = $this->talkToAPI($params, '/meetings/' . $meeting->id);
				if (!$meeting || empty($meeting->occurrences)) {
					continue;
				}

				// The original meeting has not time information
				$meeting->start_time = $meeting->occurrences[0]->start_time;
				$meeting->duration   = $meeting->occurrences[0]->duration;

				$meeting->original_id = -1;
				$meetings             = [$meeting];

				// When expanding, then also add the instances
				if ($options->get('expand', true)) {
					// Do not add the original event
					$meetings = [];
					foreach ($meeting->occurrences as $occurrence) {
						// Create the instance based on the meeting
						$instance              = $meeting;
						$instance->id          = $meeting->id . '_' . $occurrence->occurrence_id;
						$instance->original_id = $meeting->id;
						$instance->start_time  = $occurrence->start_time;
						$instance->duration    = $occurrence->duration;
						unset($instance->occurrences);
						unset($instance->recurrence);
						$instance->type = 2;

						$meetings[] = $instance;
					}
				}
			}

			// Loop over the meetings
			foreach ($meetings as $meeting) {
				$start = DPCalendarHelper::getDate($meeting->start_time);
				$end   = clone $start;
				$end->modify('+' . $meeting->duration . ' minutes');

				// Meeting start date and end date are before start date
				if ($startDateString && $startDateString > $start->format('c') && $startDateString > $start->format('c')) {
					continue;
				}

				// Meeting start date and end date are after end date
				if ($endDateString && $endDateString < $start->format('c') && $endDateString < $start->format('c')) {
					continue;
				}

				// Add the meeting
				$events[] = $this->createEventFromMeeting($meeting, $calendarId);
			}
		}

		return $events;
	}

	public function import()
	{
		try {
			$requestParams = new Registry($this->app->input->get('params', ['api_key' => null, 'api_secret' => null], 'array'));

			if (!$requestParams->get('api_key') || !$requestParams->get('api_key')) {
				throw new Exception('Invalid API params');
			}

			foreach ($this->talkToAPI($requestParams, '/users')->users as $user) {
				$model  = BaseDatabaseModel::getInstance('Extcalendar', 'DPCalendarModel');
				$params = new Registry();
				$params->set('api_key', $requestParams->get('api_key'));
				$params->set('api_secret', $requestParams->get('api_secret'));
				$params->set('user_id', $user->id);

				$table = Table::getInstance('Extcalendar', 'DPCalendarTable');
				$table->load(['title' => $user->first_name . ' ' . $user->last_name]);

				$data = [];
				if ($table->id > 0) {
					$data['id'] = $table->id;
				}
				$data['title']  = $user->first_name . ' ' . $user->last_name;
				$data['params'] = $params->toString();
				$data['plugin'] = 'zoom';

				if (!$model->save($data)) {
					$this->app->enqueueMessage($model->getError(), 'warning');
				}
			}
		} catch (Exception $e) {
			$this->app->enqueueMessage($e->getMessage(), 'warning');
		}
		$this->app->redirect(Factory::getSession()->get(
			'extcalendarOrigin',
			'index.php?option=com_dpcalendar&view=extcalendars&dpplugin=zoom&tmpl=' . $this->app->input->get('tmpl'),
			'DPCalendar'
		));
	}

	public function getRecordings(Registry $params, $meetingId, Date $start)
	{
		JLoader::import('dpcalendar.zoom.vendor.autoload', JPATH_PLUGINS);
		$accessToken = JWT::encode(['iss' => $params->get('api_key'), 'exp' => time() + 60], $params->get('api_secret'), 'HS256');
		$meetingId   = preg_replace('/[^0-9]/', '', $meetingId);

		// Fetch instances of the current meeting
		$instances = $this->talkToAPI($params, '/past_meetings/' . $meetingId . '/instances');

		// When empty, check if the meeting itself has recordings
		if (empty($instances->meetings)) {
			$instances = new stdClass();
			try {
				$instances->meetings = [$this->talkToAPI($params, '/meetings/' . $meetingId . '/recordings')];
			} catch (Exception $e) {
				// When 404 then there are no recordings available
				if ($e->getCode() != 404) {
					throw $e;
				}
			}
		}

		// No recordings found
		if (empty($instances->meetings)) {
			return [];
		}

		// Collect the recordings
		$recordingFiles = [];
		foreach ($instances->meetings as $instance) {
			// When we have multiple recordings do a time span comparison if they are on the same day
			if (count($instances->meetings) > 1 && substr($instance->start_time, 0, 10) != $start->format('Y-m-d')) {
				continue;
			}

			// Collect the recordings
			$recordings = $instance;
			if (empty($instance->recording_files)) {
				try {
					$recordings = $this->talkToAPI($params, '/meetings/' . $instance->uuid . '/recordings');
				} catch (Exception $e) {
					// When 404 then there are no recordings available
					if ($e->getCode() != 404) {
						throw $e;
					}
					$recordings->recording_files = [];
				}
			}
			foreach ($recordings->recording_files as $index => $file) {
				// Add the access token for password protected files
				$file->download_url = $file->download_url . '?access_token=' . $accessToken;
				$file->topic        = $recordings->topic;
				$recordingFiles[]   = $file;
			}
		}

		return $recordingFiles;
	}

	public function talkToAPI(Registry $params, $method, $body = '', $fields = [])
	{
		JLoader::import('dpcalendar.zoom.vendor.autoload', JPATH_PLUGINS);

		$http = new HTTP();

		$result = null;
		if ($body) {
			$result = $http->post(
				'https://api.zoom.us/v2/' . $method . '?' . http_build_query($fields),
				$body,
				JWT::encode(['iss' => $params->get('api_key'), 'exp' => time() + 60], $params->get('api_secret'), 'HS256')
			);
		} else {
			$result = $http->get(
				'https://api.zoom.us/v2/' . $method . '?' . http_build_query($fields),
				JWT::encode(['iss' => $params->get('api_key'), 'exp' => time() + 60], $params->get('api_secret'), 'HS256')
			);
		}

		if (!$result) {
			throw new Exception('No json data returned!!');
		}

		if ($result->dp->info->http_code >= 400) {
			throw new Exception($result->message, $result->dp->info->http_code);
		}

		return $result;
	}

	public function onContentPrepareForm($form, $data)
	{
		if ($form->getName() != 'com_dpcalendar.event') {
			return true;
		}

		$obj = (object)$data;
		// On save data is empty so add the form
		if (!empty($obj->id) && !is_numeric($obj->id)) {
			return true;
		}

		$form->loadFile(JPATH_PLUGINS . '/dpcalendar/zoom/forms/event.xml');

		parent::onContentPrepareForm($form, $data);
	}

	private function createEventFromMeeting($meeting, $calendarId)
	{
		$event              = $this->createEvent($meeting->id, $calendarId);
		$event->original_id = $meeting->original_id;

		$id = $meeting->id;
		if (strpos($id, '_') !== false) {
			$id = substr($id, 0, strpos($id, '_'));
		}

		$start             = DPCalendarHelper::getDate($meeting->start_time);
		$event->start_date = $start->toSql();
		$start->modify('+' . $meeting->duration . ' minutes');
		$event->end_date = $start->toSql();

		$event->title       = $meeting->topic;
		$event->description = isset($meeting->agenda) ? $meeting->agenda : '';
		$event->modified    = DPCalendarHelper::getDate($meeting->created_at)->toSql();
		$event->created     = DPCalendarHelper::getDate($meeting->created_at)->toSql();
		$event->url         = 'https://zoom.us/j/' . $id;

		$params = new Registry();
		$params->set('zoom_calendar', $calendarId);
		$params->set('zoom_meeting_nr', $id);
		$params->set('zoom_meeting_password', $meeting->password);
		$params->set('zoom_meeting_password_hashed', $meeting->encrypted_password);
		$event->params = $params->toString();

		if (empty($meeting->recurrence)) {
			return $event;
		}

		$dayMapping = [1 => 'SU', 2 => 'MO', 3 => 'TU', 4 => 'WE', 5 => 'TH', 6 => 'FR', 7 => 'SA'];

		switch ($meeting->recurrence->type) {
			case 1:
				$event->rrule = 'FREQ=DAILY;';
				break;
			case 2:
				$event->rrule = 'FREQ=WEEKLY;BYDAY=';
				$days         = array_map(function ($day) use ($dayMapping) {
					return $dayMapping[$day];
				});
				$event->rrule .= implode(',', $days, explode(',', $meeting->recurrence->weekly_days)) . ';';
				break;
			case 3:
				$event->rrule = 'FREQ=MONTHLY;';

				if (!empty($meeting->recurrence->monthly_day)) {
					$event->rrule .= 'BYMONTHDAY=' . $meeting->recurrence->monthly_day . ';';
				}

				if (!empty($meeting->recurrence->monthly_week)) {
					$event->rrule .= 'BYDAY=' . $meeting->recurrence->monthly_week . $dayMapping[$meeting->recurrence->monthly_week_day] . ';';
				}
				break;
		}

		if (!empty($meeting->recurrence->repeat_interval)) {
			$event->rrule .= 'INTERVAL=' . $meeting->recurrence->repeat_interval . ';';
		}
		if (!empty($meeting->recurrence->end_times)) {
			$event->rrule .= 'COUNT=' . $meeting->recurrence->end_times . ';';
		}
		if (!empty($meeting->recurrence->end_date_time)) {
			$event->rrule .= 'UNTIL=' . str_replace(['-', ':'], ['', ''], $meeting->recurrence->end_date_time) . ';';
		}

		return $event;
	}
}
